import os

from zrb import CmdTask, runner

from .._constant import APP_FRONTEND_DIR
from .._env import app_env_file
from ._checker import check_frontend_path
from ._group import snake_zrb_app_name_frontend_group

_CURRENT_DIR = os.path.dirname(__file__)

build_snake_zrb_app_name_frontend_once = CmdTask(
    icon="🚤",
    name="build-once",
    description="Build frontend for human readable zrb app name",
    group=snake_zrb_app_name_frontend_group,
    cwd=APP_FRONTEND_DIR,
    cmd_path=os.path.join(_CURRENT_DIR, "build-once.sh"),
    checkers=[check_frontend_path],
    env_files=[app_env_file],
)
runner.register(build_snake_zrb_app_name_frontend_once)



import os

from zrb.helper.file.text import append_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_kebab_case, to_snake_case
from zrb.task.task import Task


@typechecked
async def append_src_template_env(
    task: Task, project_dir: str, app_name: str, module_name: str
):
    kebab_app_name = to_kebab_case(app_name)
    snake_module_name = to_snake_case(module_name)
    upper_snake_module_name = snake_module_name.upper()
    src_template_env_path = os.path.join(
        project_dir, "src", kebab_app_name, "src", "template.env"
    )
    new_env_str = "\n".join(
        [
            f"APP_ENABLE_{upper_snake_module_name}_MODULE=true",
        ]
    )
    task.print_out(f"Add new environment to: {src_template_env_path}")
    await append_text_file_async(src_template_env_path, new_env_str)



import os

from zrb.helper.accessories.name import get_random_name
from zrb.task_input.str_input import StrInput

SYSTEM_USER = os.getenv("USER", "incognito")

package_name_input = StrInput(
    name="package-name",
    shortcut="p",
    description="Package name",
    prompt="Package name",
    default=get_random_name(),
)

package_description_input = StrInput(
    name="package-description",
    description="Package description",
    prompt="Package description",
    default="Just another package",
)

package_homepage_input = StrInput(
    name="package-homepage",
    description="Package homepage",
    prompt="Package homepage",
    default="".join(
        [
            "https://github.com/",
            SYSTEM_USER,
            "/{{util.to_kebab_case(input.package_name)}}",
        ]
    ),
)

package_repository_input = StrInput(
    name="package-repository",
    description="Package repository",
    prompt="Package repository",
    default="".join(
        [
            "https://github.com/",
            SYSTEM_USER,
            "/{{util.to_kebab_case(input.package_name)}}",
        ]
    ),
)

package_documentation_input = StrInput(
    name="package-documentation",
    description="Package documentation",
    prompt="Package documentation",
    default="".join(
        [
            "https://github.com/",
            SYSTEM_USER,
            "/{{util.to_kebab_case(input.package_name)}}",
        ]
    ),
)

package_author_name_input = StrInput(
    name="package-author-name",
    prompt="Package author name",
    description="Package author name",
    default=SYSTEM_USER,
)

package_author_email_input = StrInput(
    name="package-author-email",
    prompt="Package author email",
    description="Package author email",
    default=f"{SYSTEM_USER}@gmail.com",
)



import os

from zrb.helper.codemod.add_function_call import add_function_call
from zrb.helper.codemod.add_import_module import add_import_module
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_kebab_case, to_snake_case
from zrb.task.task import Task


@typechecked
async def register_module(
    task: Task, project_dir: str, app_name: str, module_name: str
):
    kebab_app_name = to_kebab_case(app_name)
    snake_module_name = to_snake_case(module_name)
    app_main_file_path = os.path.join(
        project_dir, "src", kebab_app_name, "src", "main.py"
    )
    import_module_path = ".".join(["module", snake_module_name, "register_module"])
    function_name = f"register_{snake_module_name}"
    task.print_out(f"Read code from: {app_main_file_path}")
    code = await read_text_file_async(app_main_file_path)
    task.print_out(
        f'Add import "{function_name}" from "{import_module_path}" to the code'
    )
    code = add_import_module(
        code=code, module_path=import_module_path, resource=function_name
    )
    task.print_out(f'Add "{function_name}" call to the code')
    code = add_function_call(code=code, function_name=function_name, parameters=[])
    task.print_out(f"Write modified code to: {app_main_file_path}")
    await write_text_file_async(app_main_file_path, code)



from zrb import Env, EnvFile
from zrb.helper.util import to_kebab_case, to_snake_case

from .._constant import APP_TEMPLATE_ENV_FILE_NAME, MODULES
from .._env import app_enable_otel_env


def get_service_envs(base_port: int, module_index: int, module_name: str) -> list[Env]:
    kebab_module_name = to_kebab_case(module_name)
    snake_module_name = to_snake_case(module_name)
    upper_snake_module_name = snake_module_name.upper()
    # Define service env
    app_broker_type_env = Env(
        name="APP_BROKER_TYPE",
        os_name=f"ZRB_ENV_PREFIX_{upper_snake_module_name}_APP_BROKER_TYPE",
        default="rabbitmq",
    )
    app_port_env = Env(
        name="APP_PORT",
        os_name=f"ZRB_ENV_PREFIX_{upper_snake_module_name}_APP_PORT",
        default=str(base_port + module_index + 1),
    )
    enable_module_env = Env(
        name=f"APP_ENABLE_{upper_snake_module_name}_MODULE", os_name="", default="true"
    )
    zrb_app_name_env = Env(
        name="APP_NAME",
        os_name=f"ZRB_ENV_PREFIX_{upper_snake_module_name}_APP_NAME",
        default=f"kebab-zrb-app-name-{kebab_module_name}-service",
    )
    return [
        app_broker_type_env,
        app_enable_otel_env,
        app_port_env,
        zrb_app_name_env,
        enable_module_env,
    ]


def get_service_env_file(module_name: str) -> EnvFile:
    snake_module_name = to_snake_case(module_name)
    upper_snake_module_name = snake_module_name.upper()
    return EnvFile(
        path=APP_TEMPLATE_ENV_FILE_NAME,
        prefix=f"ZRB_ENV_PREFIX_{upper_snake_module_name}",
    )


def get_disable_all_module_envs() -> list[Env]:
    disable_all_module_envs: list[Env] = []
    for module in MODULES:
        snake_module_name = to_snake_case(module)
        upper_snake_module_name = snake_module_name.upper()
        disable_all_module_envs.append(
            Env(
                name=f"APP_ENABLE_{upper_snake_module_name}_MODULE",
                os_name="",
                default="false",
            )
        )
    return disable_all_module_envs



from zrb.helper.accessories.color import colored
from zrb.helper.log import logger

logger.debug(colored("Loading zrb.task_env.constant", attrs=["dark"]))

RESERVED_ENV_NAMES = ("_ZRB_EXECUTION_ID", "_ZRB_EXECUTION_ID")



from zrb.builtin.ubuntu.install._group import ubuntu_install_group
from zrb.builtin.ubuntu.update import update_ubuntu
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask

install_ubuntu_toys = CmdTask(
    name="toys",
    group=ubuntu_install_group,
    description="Install ubuntu toy packages",
    cmd=[
        "sudo apt install -y lolcat cowsay figlet neofetch",
    ],
    retry_interval=3,
    preexec_fn=None,
)
update_and_install_toys: CmdTask = install_ubuntu_toys.copy()
update_and_install_toys.add_upstream(update_ubuntu)
runner.register(update_and_install_toys)



from integration.rpc import rpc_caller
from module.auth.component import Authorizer, RPCAuthorizer

authorizer: Authorizer = RPCAuthorizer(
    rpc_caller=rpc_caller,
    is_admin_rpc_name="auth_user_is_admin",
    is_guest_rpc_name="auth_user_is_guest",
    is_user_authorized_rpc_name="auth_is_user_authorized",
)



from abc import ABC, abstractmethod
from collections.abc import Callable, Mapping
from typing import Any, Optional

from component.serializer.serializer import JsonSerializer, Serializer

TEventHandler = Callable[[Any], Any]


class Admin(ABC):
    @abstractmethod
    async def create_events(self, event_names: list[str]):
        pass

    @abstractmethod
    async def delete_events(self, event_names: list[str]):
        pass


class Publisher(ABC):
    @abstractmethod
    async def publish(self, event_name: str, message: Any):
        pass


class Consumer(ABC):
    @abstractmethod
    def register(self, event_name: str) -> Callable[[TEventHandler], Any]:
        pass

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass


class MessageSerializer:
    def __init__(self, serializers: Optional[Mapping[str, Serializer]] = None):
        serializers = serializers if serializers is not None else {}
        self.serializers: Mapping[str, Serializer] = serializers
        self.default_serializer = JsonSerializer()

    def encode(self, event_name: str, message: Any) -> Any:
        serializer = self._get_serializer(event_name)
        return serializer.encode(message)

    def decode(self, event_name: str, encoded_message: Any) -> Any:
        serializer = self._get_serializer(event_name)
        return serializer.decode(encoded_message)

    def _get_serializer(self, event_name: str) -> Serializer:
        return self.serializers.get(event_name, self.default_serializer)


def must_get_message_serializer(
    serializer: Optional[MessageSerializer] = None,
) -> MessageSerializer:
    if serializer is None:
        return MessageSerializer()
    return serializer



import os
from typing import Any

from zrb import Task, python_task, runner
from zrb.helper.accessories.color import colored

from ..._project import start_project
from .._constant import PREFER_MICROSERVICES
from .._input import host_input, https_input, local_input
from ..container._input import enable_monitoring_input
from ._group import snake_zrb_app_name_microservices_group
from .start_gateway import start_snake_zrb_app_name_gateway
from .start_services import start_snake_zrb_app_name_services

_CURRENT_DIR = os.path.dirname(__file__)


@python_task(
    icon="🚤",
    name="start",
    description="Start human readable zrb app name as a microservices",
    group=snake_zrb_app_name_microservices_group,
    inputs=[
        local_input,
        enable_monitoring_input,
        host_input,
        https_input,
    ],
    upstreams=[
        start_snake_zrb_app_name_gateway,
        *start_snake_zrb_app_name_services,
    ],
)
def start_snake_zrb_app_name_microservices(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    task.print_out(colored("Microservices started", color="yellow"))


if PREFER_MICROSERVICES:
    start_snake_zrb_app_name_microservices >> start_project

runner.register(start_snake_zrb_app_name_microservices)



from integration.db_connection import engine
from integration.log import logger
from module.auth.entity.permission.repo import PermissionDBRepo, PermissionRepo

permission_repo: PermissionRepo = PermissionDBRepo(logger=logger, engine=engine)



import os
from typing import Any

from zrb.builtin.project._group import project_group
from zrb.builtin.project._input import project_dir_input
from zrb.builtin.project.add.project_task import add_project_tasks
from zrb.builtin.project.create._helper import copy_resource_replacement_mutator
from zrb.builtin.project.create._input import (
    project_author_email_input,
    project_author_name_input,
    project_description_input,
    project_name_input,
)
from zrb.config.config import VERSION
from zrb.helper.accessories.color import colored
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask
from zrb.task.decorator import python_task
from zrb.task.resource_maker import ResourceMaker
from zrb.task.task import Task

_CURRENT_DIR = os.path.dirname(__file__)
_SHELL_SCRIPTS_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(_CURRENT_DIR))), "shell-scripts"
)
_SYSTEM_USER = os.getenv("USER", "incognito")
_IMAGE_DEFAULT_NAMESPACE = os.getenv(
    "PROJECT_IMAGE_DEFAULT_NAMESPACE", "docker.io/" + _SYSTEM_USER
)


@python_task(
    name="validate",
    inputs=[project_dir_input],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get("project_dir")
    if os.path.isfile(os.path.join(project_dir, "zrb_init.py")):
        raise Exception(f"Project directory already exists: {project_dir}")


copy_resource = ResourceMaker(
    name="copy-resource",
    inputs=[
        project_dir_input,
        project_name_input,
        project_description_input,
        project_author_name_input,
        project_author_email_input,
    ],
    upstreams=[validate],
    replacements={
        "zrbProjectDir": "{{input.project_dir}}",
        "zrbProjectName": "{{input.project_name}}",
        "zrbProjectDescription": "{{input.project_description}}",
        "zrbProjectAuthorName": "{{input.project_author_name}}",
        "zrbProjectAuthorEmail": "{{input.project_author_email}}",
        "zrbImageDefaultNamespace": _IMAGE_DEFAULT_NAMESPACE,
        "zrbVersion": VERSION,
    },
    replacement_mutator=copy_resource_replacement_mutator,
    template_path=os.path.join(_CURRENT_DIR, "template"),
    destination_path="{{input.project_dir}}",
    excludes=[
        "*/__pycache__",
    ],
)

_add_project_tasks = add_project_tasks.copy()
_add_project_tasks.add_upstream(copy_resource)

init_git = CmdTask(
    name="init-git",
    upstreams=[copy_resource],
    inputs=[project_dir_input],
    cmd_path=[
        os.path.join(_SHELL_SCRIPTS_DIR, "_common-util.sh"),
        os.path.join(_CURRENT_DIR, "init-git.sh"),
    ],
)


@python_task(
    name="create",
    description="Create a new project",
    inputs=[
        project_dir_input,
        project_name_input,
        project_description_input,
        project_author_name_input,
        project_author_email_input,
    ],
    group=project_group,
    upstreams=[_add_project_tasks, init_git],
)
def create_project(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    task.print_out(colored("Project created", color="yellow"))
    task.print_out(colored("Happy coding :)", color="yellow"))


runner.register(create_project)



import os

from zrb.helper.codemod.add_property_to_class import add_property_to_class
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_pascal_case, to_snake_case
from zrb.task.task import Task

from ._common import get_app_dir, get_sqlalchemy_column_type


@typechecked
async def inject_repo(
    task: Task,
    project_dir: str,
    app_name: str,
    module_name: str,
    entity_name: str,
    column_name: str,
    column_type: str,
):
    snake_module_name = to_snake_case(module_name)
    snake_entity_name = to_snake_case(entity_name)
    pascal_entity_name = to_pascal_case(entity_name)
    snake_column_name = to_snake_case(column_name)
    repo_file_path = os.path.join(
        get_app_dir(project_dir, app_name),
        "src",
        "module",
        snake_module_name,
        "entity",
        snake_entity_name,
        "repo.py",
    )
    task.print_out(f"Read code from: {repo_file_path}")
    code = await read_text_file_async(repo_file_path)
    task.print_out(f'Add column "{snake_column_name}" to the repo')
    sqlalchemy_column_type = get_sqlalchemy_column_type(column_type)
    code = add_property_to_class(
        code=code,
        class_name=f"DBEntity{pascal_entity_name}",
        property_name=snake_column_name,
        property_type="Column",
        property_value=f"Column({sqlalchemy_column_type})",
    )
    task.print_out(f"Write modified code to: {repo_file_path}")
    await write_text_file_async(repo_file_path, code)



from zrb.task_input.int_input import IntInput
from zrb.task_input.str_input import StrInput

http_port_input = IntInput(
    name="http-port",
    shortcut="p",
    description="HTTP port",
    prompt="HTTP port",
    default=8080,
)

compose_command_input = StrInput(
    name="compose-command",
    shortcut="c",
    description="Compose command",
    prompt="Compose command",
    default="up",
)



from ._group import snake_zrb_app_name_monolith_group
from .deploy import deploy_snake_zrb_app_name_monolith
from .start import start_snake_zrb_app_name_monolith

assert snake_zrb_app_name_monolith_group
assert deploy_snake_zrb_app_name_monolith
assert start_snake_zrb_app_name_monolith



from zrb.helper.accessories.color import colored
from zrb.helper.advertisement import Advertisement
from zrb.helper.log import logger

logger.debug(colored("Loading zrb.advertisement", attrs=["dark"]))

# flake8: noqa E501

advertisements: list[Advertisement] = [
    Advertisement(
        content="\n".join(
            [
                "Support zrb growth and development!",
                "☕ Donate at: https://stalchmst.com/donation",
                "🐙 Submit issues/PR at: https://github.com/state-alchemists/zrb",
                "🐤 Follow us at: https://twitter.com/zarubastalchmst",
            ]
        ),
        time_pattern=".*",
    ),
]



from collections.abc import Callable
from typing import Optional

from zrb.helper.accessories.color import colored
from zrb.helper.callable import run_async
from zrb.helper.log import logger

logger.debug(colored("Loading zrb.task.looper", attrs=["dark"]))


class Looper:
    def __init__(self):
        self._queue: Mapping[str, list[Optional[bool]]] = {}
        self._should_stop = False

    async def pop(self, identifier: str) -> Optional[bool]:
        if identifier not in self._queue or len(self._queue[identifier]) == 0:
            return None
        return self._queue[identifier].pop(0)

    def stop(self):
        self._should_stop = True

    def is_registered(self, identifier: str) -> bool:
        return identifier in self._queue

    async def register(self, identifier: str, function: Callable[..., Optional[bool]]):
        if identifier in self._queue:
            return
        self._queue[identifier] = []
        while not self._should_stop:
            try:
                result = await run_async(function)
                if result is not None:
                    if not result:
                        continue
                    while len(self._queue[identifier]) > 1000:
                        self._queue[identifier].pop(0)
                    self._queue[identifier].append(result)
            except KeyboardInterrupt:
                self.stop()
                break


looper = Looper()



import time

from config import (
    APP_DB_CONNECTION,
    APP_DB_ENGINE_SHOW_LOG,
    APP_DB_POOL_MAX_OVERFLOW,
    APP_DB_POOL_PRE_PING,
    APP_DB_POOL_SIZE,
)
from integration.log import logger
from sqlalchemy import Engine, create_engine, event


@event.listens_for(Engine, "before_cursor_execute")
def before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    context._query_start_time = time.time()
    one_line_statement = " ".join(statement.split("\n"))
    logger.info(f"💽 [start query] SQL:`{one_line_statement}`, parameters {parameters}")


@event.listens_for(Engine, "after_cursor_execute")
def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    total = time.time() - context._query_start_time
    one_line_statement = " ".join(statement.split("\n"))
    logger.info(
        " ".join(
            [
                f"💽 [query completed] Elapsed time: {total} ms,",
                f"SQL: `{one_line_statement}`,",
                f"parameters {parameters}",
            ]
        )
    )


engine: Engine = create_engine(
    APP_DB_CONNECTION,
    echo=APP_DB_ENGINE_SHOW_LOG,
    pool_pre_ping=APP_DB_POOL_PRE_PING,
    pool_size=APP_DB_POOL_SIZE,
    max_identifier_length=APP_DB_POOL_MAX_OVERFLOW,
)



from config import (
    APP_AUTH_ACCESS_TOKEN_EXPIRE_SECONDS,
    APP_AUTH_ADMIN_ACTIVE,
    APP_AUTH_REFRESH_TOKEN_EXPIRE_SECONDS,
)
from integration.messagebus import publisher
from module.auth.entity.user.model import UserModel
from module.auth.integration import access_token_util, refresh_token_util
from module.auth.integration.model.permission_model import permission_model
from module.auth.integration.repo.user_repo import user_repo
from module.auth.integration.user import admin_user, admin_user_password, guest_user

user_model: UserModel = UserModel(
    repo=user_repo,
    publisher=publisher,
    permission_model=permission_model,
    access_token_util=access_token_util,
    access_token_expire_seconds=APP_AUTH_ACCESS_TOKEN_EXPIRE_SECONDS,
    refresh_token_util=refresh_token_util,
    refresh_token_expire_seconds=APP_AUTH_REFRESH_TOKEN_EXPIRE_SECONDS,
    guest_user=guest_user,
    admin_user=admin_user if APP_AUTH_ADMIN_ACTIVE else None,
    admin_user_password=admin_user_password,
)



from zrb.builtin.ubuntu.install._group import ubuntu_install_group
from zrb.builtin.ubuntu.install.all import install_ubuntu_all
from zrb.builtin.ubuntu.install.essentials import install_ubuntu_essentials
from zrb.builtin.ubuntu.install.tex import install_ubuntu_tex
from zrb.builtin.ubuntu.install.toys import install_ubuntu_toys

assert ubuntu_install_group
assert install_ubuntu_all
assert install_ubuntu_essentials
assert install_ubuntu_tex
assert install_ubuntu_toys



from zrb.builtin.process._group import process_group
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask
from zrb.task_input.str_input import StrInput

get_process_pid_by_name = CmdTask(
    name="get-pid-by-name",
    group=process_group,
    description="Get PID by name",
    inputs=[
        StrInput(
            name="name",
            shortcut="n",
            default="python",
            description="Process name to be checked",
            prompt="Process name to be checked",
        )
    ],
    cmd='pgrep "{{ input.name }}" || exit 0',
    checking_interval=3,
)
runner.register(get_process_pid_by_name)



from ._group import project_image_group
from .build import build_project_images
from .push import push_project_images

assert project_image_group
assert build_project_images
assert push_project_images



from collections.abc import Callable, Mapping
from typing import Any, Optional, Union

from zrb.config.config import SHOW_PROMPT
from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.string.conversion import to_variable_name
from zrb.helper.string.jinja import is_probably_jinja
from zrb.helper.typecheck import typechecked
from zrb.task_input.any_input import AnyInput
from zrb.task_input.constant import RESERVED_INPUT_NAMES

logger.debug(colored("Loading zrb.task_input.base_input", attrs=["dark"]))

# flake8: noqa E501

InputCallback = Callable[[Mapping[str, Any], Any], Any]
InputDefault = Callable[[Mapping[str, Any]], Any]


@typechecked
class BaseInput(AnyInput):
    """
    Base class for all Input.

    Input is an interface for a Task to read user input at the beginning of the execution.

    Attributes:
        name (str): The name of the input, used as a unique identifier.
        shortcut (Optional[str]): An optional single-character shortcut for the input.
        default (Optional[Any]): The default value of the input.
        callback (Optional[Any]): The default value of the input.
        description (Optional[str]): A brief description of what the input is for.
        show_default (Union[bool, JinjaTemplate, None]): Determines the default value to be shown.
        prompt (Union[bool, str]): The prompt text to be displayed when asking for the input.
        confirmation_prompt (Union[bool, str]): A prompt for confirmation if required.
        prompt_required (bool): Indicates whether a prompt is required.
        hide_input (bool): If True, the input value will be hidden (e.g., for passwords).
        is_flag (Optional[bool]): Specifies whether the input is a flag.
        flag_value (Optional[Any]): The value to be used if the input is a flag.
        multiple (bool): Allows multiple values for this input if True.
        count (bool): If True, counts the occurrences of the input.
        allow_from_autoenv (bool): If True, allows values to be automatically sourced from the environment.
        type (Optional[Any]): The expected type of the input value.
        hidden (bool): If True, the input is hidden and not rendered.
        show_choices (bool): Indicates whether to show available choices for the input.
        show_envvar (bool): If True, shows the corresponding environment variable.
        nargs (int): Number of arguments expected for this input.
        should_render (bool): Determines whether the input should be rendered.

    Examples:
        >>> from zrb import Input, Task
        >>> task = Task(
        >>>     name='task',
        >>>     inputs=[
        >>>         Input(name='delay', default=10, description='Delay')
        >>>     ]
        >>> )
    """

    __value_cache: Mapping[str, Any] = {}
    __default_cache: Mapping[str, Any] = {}

    def __init__(
        self,
        name: str,
        shortcut: Optional[str] = None,
        default: Optional[Union[Any, InputDefault]] = None,
        callback: Optional[InputCallback] = None,
        description: Optional[str] = None,
        show_default: Union[bool, str] = True,
        prompt: Union[bool, str] = True,
        confirmation_prompt: Union[bool, str] = False,
        prompt_required: bool = True,
        hide_input: bool = False,
        is_flag: Optional[bool] = None,
        flag_value: Optional[Any] = None,
        multiple: bool = False,
        count: bool = False,
        allow_from_autoenv: bool = True,
        type: Optional[Any] = None,
        hidden: bool = False,
        show_choices: bool = True,
        show_envvar: bool = False,
        nargs: int = 1,
        should_render: bool = True,
    ):
        if name in RESERVED_INPUT_NAMES:
            raise ValueError(f"Forbidden input name: {name}")
        self._name = name
        self._shortcut = shortcut
        self._prompt = prompt
        self._default = default
        self._callback = callback
        self._help = description if description is not None else name
        self._type = type
        self._show_default = show_default
        self._confirmation_prompt = confirmation_prompt
        self._prompt_required = prompt_required
        self._hide_input = hide_input
        self._is_flag = is_flag
        self._flag_value = flag_value
        self._multiple = multiple
        self._count = count
        self._allow_from_autoenv = allow_from_autoenv
        self._hidden = hidden
        self._show_choices = show_choices
        self._show_envvar = show_envvar
        self._nargs = nargs
        self.__should_render = should_render

    def get_name(self) -> str:
        return self._name

    def get_default(self) -> Any:
        if callable(self._default):
            return self._default(self.__value_cache)
        return self._default

    def get_param_decl(self) -> list[str]:
        param_decl: list[str] = [f"--{self._name}"]
        if self._shortcut is not None:
            param_decl.append(f"-{self._shortcut}")
        return param_decl

    def get_options(self) -> Mapping[str, Any]:
        options: Mapping[str, Any] = {
            "help": self._help,
            "type": self._type,
            "show_default": self._get_calculated_show_default(),
            "confirmation_prompt": self._confirmation_prompt,
            "prompt_required": self._prompt_required,
            "hide_input": self._hide_input,
            "is_flag": self._is_flag,
            "flag_value": self._flag_value,
            "multiple": self._multiple,
            "count": self._count,
            "allow_from_autoenv": self._allow_from_autoenv,
            "hidden": self._hidden,
            "show_choices": self._show_choices,
            "show_envvar": self._show_envvar,
            "nargs": self._nargs,
            "callback": self._wrapped_callback,
            "default": self._wrapped_default,
        }
        if SHOW_PROMPT:
            options["prompt"] = self._prompt
        return options

    def _wrapped_callback(self, ctx, param, value) -> Any:
        var_name = to_variable_name(self.get_name())
        if var_name not in self.__value_cache:
            if callable(self._callback):
                result = self._callback(self.__value_cache, value)
                self.__value_cache[var_name] = result
                return result
            self.__value_cache[var_name] = value
            return value
        return self.__value_cache[var_name]

    def _wrapped_default(self) -> Any:
        var_name = to_variable_name(self.get_name())
        if var_name not in self.__default_cache:
            if callable(self._default):
                default = self._default(self.__value_cache)
                self.__default_cache[var_name] = default
                return default
            self.__default_cache[var_name] = self._default
            return self._default
        return self.__default_cache[var_name]

    def _get_calculated_show_default(self) -> str:
        if self._show_default != True:
            return self._show_default
        if callable(self._default) or (
            self.__should_render and is_probably_jinja(self._default)
        ):
            return colored("Autogenerated", color="yellow")
        return f"{self._default}"

    def should_render(self) -> bool:
        return self.__should_render

    def is_hidden(self) -> bool:
        return False



from ._group import snake_zrb_app_name_group
from .backend import (
    prepare_snake_zrb_app_name_backend,
    snake_zrb_app_name_backend_group,
)
from .container import (
    remove_snake_zrb_app_name_container,
    snake_zrb_app_name_container_group,
    snake_zrb_app_name_microservices_container_group,
    snake_zrb_app_name_monolith_container_group,
    snake_zrb_app_name_support_container_group,
    start_snake_zrb_app_name_microservices_container,
    start_snake_zrb_app_name_monolith_container,
    start_snake_zrb_app_name_support_container,
    stop_snake_zrb_app_name_container,
)
from .destroy import destroy_snake_zrb_app_name
from .frontend import (
    build_snake_zrb_app_name_frontend,
    build_snake_zrb_app_name_frontend_once,
    snake_zrb_app_name_frontend_group,
)
from .image import (
    build_snake_zrb_app_name_image,
    push_snake_zrb_app_name_image,
    snake_zrb_app_name_image_group,
)
from .load_test import load_test_snake_zrb_app_name
from .microservices import (
    deploy_snake_zrb_app_name_microservices,
    snake_zrb_app_name_microservices_group,
    start_snake_zrb_app_name_microservices,
)
from .monolith import (
    deploy_snake_zrb_app_name_monolith,
    snake_zrb_app_name_monolith_group,
    start_snake_zrb_app_name_monolith,
)
from .test import test_snake_zrb_app_name

assert snake_zrb_app_name_group
assert snake_zrb_app_name_backend_group
assert snake_zrb_app_name_frontend_group
assert snake_zrb_app_name_microservices_group
assert snake_zrb_app_name_monolith_group
assert snake_zrb_app_name_image_group
assert snake_zrb_app_name_container_group
assert snake_zrb_app_name_microservices_container_group
assert snake_zrb_app_name_monolith_container_group
assert snake_zrb_app_name_support_container_group
assert destroy_snake_zrb_app_name
assert start_snake_zrb_app_name_microservices
assert deploy_snake_zrb_app_name_microservices
assert start_snake_zrb_app_name_monolith
assert deploy_snake_zrb_app_name_monolith
assert remove_snake_zrb_app_name_container
assert stop_snake_zrb_app_name_container
assert start_snake_zrb_app_name_microservices_container
assert start_snake_zrb_app_name_monolith_container
assert start_snake_zrb_app_name_support_container
assert build_snake_zrb_app_name_image
assert push_snake_zrb_app_name_image
assert build_snake_zrb_app_name_frontend_once
assert build_snake_zrb_app_name_frontend
assert prepare_snake_zrb_app_name_backend
assert load_test_snake_zrb_app_name
assert test_snake_zrb_app_name



import os
from collections.abc import Callable
from typing import Any

from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.task.task import Task


@typechecked
def write_config(
    template_file: str, config_file: str, remove_old_config: bool = False
) -> Callable[..., Any]:
    async def set_config(*args, **kwargs):
        task: Task = kwargs.get("_task")
        rendered_config_file = os.path.expandvars(
            os.path.expanduser(task.render_str(config_file))
        )
        rendered_template_file = os.path.expandvars(
            os.path.expanduser(task.render_str(template_file))
        )
        if remove_old_config and os.path.exists(rendered_config_file):
            task.print_out(f"Removing {rendered_config_file}")
            os.remove(rendered_config_file)
        additional_content = await read_text_file_async(rendered_template_file)
        content = ""
        if os.path.exists(rendered_config_file):
            content = await read_text_file_async(rendered_config_file) + "\n"
        new_content = content + additional_content
        task.print_out(f"Writing content to {rendered_config_file}")
        await write_text_file_async(rendered_config_file, new_content)

    return set_config



from collections.abc import Callable
from typing import Any

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.task.any_task import AnyTask

logger.debug(colored("Loading zrb.task.any_task_event_handler", attrs=["dark"]))

_TaskEventHandler = Callable[[AnyTask], Any]
OnTriggered = _TaskEventHandler
OnWaiting = _TaskEventHandler
OnSkipped = _TaskEventHandler
OnStarted = _TaskEventHandler
OnReady = _TaskEventHandler
OnRetry = _TaskEventHandler
OnFailed = Callable[[AnyTask, bool, Exception], Any]



from abc import ABC, abstractmethod
from typing import TypeVar, Union

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask

logger.debug(colored("Loading zrb.task.parallel", attrs=["dark"]))

TParallel = TypeVar("TParallel", bound="Parallel")


class AnyParallel(ABC):
    @abstractmethod
    def get_tasks(self) -> list[AnyTask]:
        pass


@typechecked
class Parallel(AnyParallel):
    def __init__(self, *tasks: AnyTask):
        self.__tasks = list(tasks)

    def get_tasks(self) -> list[AnyTask]:
        return self.__tasks

    def __rshift__(
        self, operand: Union[AnyTask, AnyParallel]
    ) -> Union[AnyTask, AnyParallel]:
        if isinstance(operand, AnyTask):
            for task in self.__tasks:
                operand.add_upstream(task)
            return operand
        if isinstance(operand, AnyParallel):
            other_tasks: list[AnyTask] = operand.get_tasks()
            for task in self.__tasks:
                for other_task in other_tasks:
                    other_task.add_upstream(task)
            return operand



import os

from zrb.task_input.str_input import StrInput

_PROJECT_DIR = os.getenv("ZRB_PROJECT_DIR", ".")

project_dir_input = StrInput(
    name="project-dir",
    shortcut="d",
    description="Project directory",
    prompt="Project directory",
    default=_PROJECT_DIR,
)



from zrb.builtin.devtool.install._group import devtool_install_group
from zrb.builtin.devtool.install.aws.aws import install_aws
from zrb.builtin.devtool.install.docker.docker import install_docker
from zrb.builtin.devtool.install.gcloud.gcloud import install_gcloud
from zrb.builtin.devtool.install.gvm.gvm import install_gvm
from zrb.builtin.devtool.install.helix.helix import install_helix
from zrb.builtin.devtool.install.helm.helm import install_helm
from zrb.builtin.devtool.install.kubectl.kubectl import install_kubectl
from zrb.builtin.devtool.install.nvm.nvm import install_nvm
from zrb.builtin.devtool.install.pulumi.pulumi import install_pulumi
from zrb.builtin.devtool.install.pyenv.pyenv import install_pyenv
from zrb.builtin.devtool.install.sdkman.sdkman import install_sdkman
from zrb.builtin.devtool.install.selenium.selenium import install_selenium
from zrb.builtin.devtool.install.terraform.terraform import install_terraform
from zrb.builtin.devtool.install.tmux.tmux import install_tmux
from zrb.builtin.devtool.install.zsh.zsh import install_zsh

assert devtool_install_group
assert install_aws
assert install_docker
assert install_gcloud
assert install_gvm
assert install_helix
assert install_helm
assert install_kubectl
assert install_nvm
assert install_pulumi
assert install_pyenv
assert install_sdkman
assert install_selenium
assert install_terraform
assert install_tmux
assert install_zsh



import os
from contextlib import asynccontextmanager

from config import (
    APP_DB_AUTO_MIGRATE,
    APP_ENABLE_EVENT_HANDLER,
    APP_ENABLE_FRONTEND,
    APP_ENABLE_RPC_SERVER,
    APP_NAME,
    APP_SRC_DIR,
)
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from helper.async_task import create_task
from integration.app.app_state import app_state, set_not_ready_on_error
from integration.log import logger
from integration.messagebus import consumer
from integration.rpc import rpc_server
from migrate import migrate


@asynccontextmanager
async def app_lifespan(app: FastAPI):
    logger.info(f"{APP_NAME} started")
    if APP_DB_AUTO_MIGRATE:
        await migrate()
    app_state.set_liveness(True)
    if APP_ENABLE_EVENT_HANDLER:
        create_task(consumer.start(), on_error=set_not_ready_on_error)
    if APP_ENABLE_RPC_SERVER:
        create_task(rpc_server.start(), on_error=set_not_ready_on_error)
    if APP_ENABLE_FRONTEND:
        build_path = os.path.join(APP_SRC_DIR, "frontend", "build")
        app.mount(
            path="",
            app=StaticFiles(directory=build_path, html=True),
            name="frontend-static-resources",
        )
    app_state.set_readiness(True)
    logger.info(f"{APP_NAME} started")
    yield
    if APP_ENABLE_EVENT_HANDLER:
        await consumer.stop()
    logger.info(f"{APP_NAME} closed")



import os

from zrb import CmdTask, Env

from .._checker import app_local_checker
from .._constant import APP_DIR, AUTOMATE_DIR
from .._env import app_enable_otel_env, app_env_file
from .._input import host_input, https_input, local_input
from ..backend import prepare_snake_zrb_app_name_backend
from ..container._input import enable_monitoring_input
from ..container.support import start_snake_zrb_app_name_support_container
from ..frontend import build_snake_zrb_app_name_frontend

_CURRENT_DIR = os.path.dirname(__file__)

start_snake_zrb_app_name_gateway = CmdTask(
    icon="🚪",
    name="start-kebab-zrb-app-name-gateway",
    inputs=[
        local_input,
        enable_monitoring_input,
        host_input,
        https_input,
    ],
    upstreams=[
        start_snake_zrb_app_name_support_container,
        build_snake_zrb_app_name_frontend,
        prepare_snake_zrb_app_name_backend,
    ],
    cwd=APP_DIR,
    env_files=[
        app_env_file,
    ],
    envs=[
        Env(name="APP_DB_AUTO_MIGRATE", default="false", os_name=""),
        Env(name="APP_ENABLE_EVENT_HANDLER", default="false", os_name=""),
        Env(name="APP_ENABLE_RPC_SERVER", default="false", os_name=""),
        app_enable_otel_env,
    ],
    cmd_path=[
        os.path.join(AUTOMATE_DIR, "activate-venv.sh"),
        os.path.join(_CURRENT_DIR, "start.sh"),
    ],
    checkers=[
        app_local_checker,
    ],
)



from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar

from component.repo.search_filter import SearchFilter
from pydantic import BaseModel

SchemaData = TypeVar("SchemaData", bound=BaseModel)
Schema = TypeVar("Schema", bound=BaseModel)


class Repo(Generic[Schema, SchemaData], ABC):
    @abstractmethod
    async def get_by_id(self, id: str) -> Schema:
        pass

    @abstractmethod
    async def get(
        self, search_filter: Optional[SearchFilter], limit: int, offset: int
    ) -> list[Schema]:
        pass

    @abstractmethod
    async def count(self, search_filter: Optional[SearchFilter]) -> int:
        pass

    @abstractmethod
    async def insert(self, data: SchemaData) -> Schema:
        pass

    @abstractmethod
    async def update(self, id: str, data: SchemaData) -> Schema:
        pass

    @abstractmethod
    async def delete(self, id: str) -> Schema:
        pass



from zrb.helper.accessories.color import colored
from zrb.helper.log import logger

logger.debug(colored("Loading zrb.task_input.constant", attrs=["dark"]))

RESERVED_INPUT_NAMES = ("_task", "_args")



from zrb.builtin._helper.reccuring_action import create_recurring_action
from zrb.runner import runner
from zrb.task.server import Controller, Server
from zrb.task.time_watcher import TimeWatcher
from zrb.task_input.str_input import StrInput

schedule = Server(
    name="schedule",
    icon="📅",
    color="yellow",
    description="Show message/run command periodically",
    inputs=[
        StrInput(
            name="schedule",
            default="* * * * *",
            prompt="Schedule cron pattern (minute hour day(month) month day(week)",  # noqa
            description="Schedule cron pattern to show the message",
        ),
    ],
    controllers=[
        Controller(
            name="periodic",
            trigger=TimeWatcher(
                name="watch-time",
                color="cyan",
                icon="⏰",
                schedule="{{input.schedule}}",
            ),
            action=create_recurring_action(
                notif_title="Schedule",
            ),
        )
    ],
)
runner.register(schedule)



import os

from zrb import BoolInput, IntInput, StrInput

local_input = BoolInput(
    name="local-kebab-zrb-app-name",
    description='Use "kebab-zrb-app-name" on local machine',
    prompt='Use "kebab-zrb-app-name" on local machine?',
    default=True,
)

https_input = BoolInput(
    name="kebab-zrb-app-name-https",
    description='Whether "kebab-zrb-app-name" run on HTTPS',
    prompt='Is "kebab-zrb-app-name" run on HTTPS?',
    default=False,
)

host_input = StrInput(
    name="kebab-zrb-app-name-host",
    description='Hostname of "kebab-zrb-app-name"',
    prompt='Hostname of "kebab-zrb-app-name"',
    default="localhost",
)

replica_input = IntInput(
    name="kebab-zrb-app-name-replica",
    description='Replica of "kebab-zrb-app-name"',
    prompt='Replica of "kebab-zrb-app-name"',
    default=1,
)

pulumi_stack_input = StrInput(
    name="kebab-zrb-app-name-pulumi-stack",
    description='Pulumi stack name for "kebab-zrb-app-name"',
    prompt='Pulumi stack name for "kebab-zrb-app-name"',
    default=os.getenv("ZRB_ENV", "dev"),
)



import subprocess
from collections.abc import Mapping

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.git.detect_changes", attrs=["dark"]))


@typechecked
class ModificationState:
    def __init__(self):
        self.plus: bool = False
        self.minus: bool = False


@typechecked
def get_modified_file_states(commit: str) -> Mapping[str, ModificationState]:
    exit_status, output = subprocess.getstatusoutput(f"git show {commit}")
    if exit_status != 0:
        raise Exception(output)
    lines = output.split("\n")
    modified_files: Mapping[str, ModificationState] = {}
    for line in lines:
        if not line.startswith("---") and not line.startswith("+++"):
            continue
        if line[4:6] != "a/" and line[4:6] != "b/":
            continue
        # line should contains something like `--- a/some-file.txt`
        file = line[6:]
        if file not in modified_files:
            modified_files[file] = ModificationState()
        modification_state = modified_files[file]
        if line.startswith("---"):
            modification_state.minus = True
        if line.startswith("+++"):
            modification_state.plus = True
    return modified_files



from zrb.builtin.process.pid._group import process_pid_group
from zrb.builtin.process.pid.get_by_name import get_process_pid_by_name
from zrb.builtin.process.pid.get_by_port import get_process_pid_by_port

assert process_pid_group
assert get_process_pid_by_name
assert get_process_pid_by_port



import os
import traceback
from functools import lru_cache

import click

from zrb.config.config import INIT_MODULES, INIT_SCRIPTS, SHOULD_LOAD_BUILTIN, VERSION
from zrb.helper.accessories.color import colored
from zrb.helper.loader.load_module import load_module
from zrb.helper.loader.load_script import load_script
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.runner import runner

logger.debug(colored("Loading zrb.helper.cli", attrs=["dark"]))

HELP = f"""
                bb
   zzzzz rr rr  bb
     zz  rrr  r bbbbbb
    zz   rr     bb   bb
   zzzzz rr     bbbbbb   {VERSION}
   _ _ . .  . _ .  _ . . .

Super framework for your super app.

☕ Donate at: https://stalchmst.com/donation
🐙 Submit issues/PR at: https://github.com/state-alchemists/zrb
🐤 Follow us at: https://twitter.com/zarubastalchmst
"""


class MultilineHelpClickGroup(click.Group):
    def format_help_text(self, ctx, formatter):
        formatter.write(self.help)


@lru_cache
@typechecked
def create_cli() -> click.Group:
    logger.info(colored("Prepare CLI", attrs=["dark"]))
    zrb_cli_group = MultilineHelpClickGroup(name="zrb", help=HELP)
    # Load default tasks
    if SHOULD_LOAD_BUILTIN:
        logger.info(colored("Load builtins", attrs=["dark"]))
        from zrb import builtin

        assert builtin
    # load from ZRB_INIT_MODULES
    for init_module in INIT_MODULES:
        try:
            load_module(init_module)
        except Exception:
            logger.error(
                colored(
                    f"Failed to load module {init_module}",
                    color="red",
                    attrs=["bold"],
                )
            )
    # Load zrb_init.py
    project_dir = os.getenv("ZRB_PROJECT_DIR", os.getcwd())
    project_script = os.path.join(project_dir, "zrb_init.py")
    load_script(script_path=project_script, sys_path_index=0)
    # load from ZRB_INIT_SCRIPTS environment
    for index, init_script in enumerate(INIT_SCRIPTS):
        logger.info(colored(f"Load module from {init_script}", attrs=["dark"]))
        try:
            load_script(script_path=init_script, sys_path_index=index + 1)
        except Exception:
            logger.error(
                colored(
                    f"Failed to load module from {init_script}",
                    color="red",
                    attrs=["bold"],
                )
            )
            traceback.print_exc()
    # Serve all tasks registered to runner
    logger.info(colored("Serve CLI", attrs=["dark"]))
    cli = runner.serve(zrb_cli_group)
    return cli



import os
import pathlib
from collections.abc import Callable, Iterable
from typing import Any, Optional, Union

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import JinjaTemplate
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.base_remote_cmd_task import BaseRemoteCmdTask, RemoteConfig
from zrb.task.cmd_task import CmdTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.rsync_task", attrs=["dark"]))

_CURRENT_DIR = os.path.dirname(__file__)
_SHELL_SCRIPT_DIR = os.path.join(_CURRENT_DIR, "..", "shell-scripts")
with open(os.path.join(_SHELL_SCRIPT_DIR, "rsync-util.sh")) as file:
    _RSYNC_UTIL_SCRIPT = file.read()

ensure_rsync_is_installed = CmdTask(
    name="ensure-ssh-is-installed",
    cmd_path=[
        os.path.join(_SHELL_SCRIPT_DIR, "_common-util.sh"),
        os.path.join(_SHELL_SCRIPT_DIR, "ensure-ssh-is-installed.sh"),
        os.path.join(_SHELL_SCRIPT_DIR, "ensure-rsync-is-installed.sh"),
    ],
    preexec_fn=None,
)


@typechecked
class RsyncTask(BaseRemoteCmdTask):
    def __init__(
        self,
        name: str,
        remote_configs: Iterable[RemoteConfig],
        src: JinjaTemplate,
        dst: JinjaTemplate,
        src_is_remote: bool = False,
        dst_is_remote: bool = True,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        executable: Optional[str] = None,
        cwd: Optional[Union[str, pathlib.Path]] = None,
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0.05,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        preexec_fn: Optional[Callable[[], Any]] = os.setsid,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
    ):
        parsed_src = self._get_parsed_path(src_is_remote, src)
        parsed_dst = self._get_parsed_path(dst_is_remote, dst)
        cmd = f'auth_rsync "{parsed_src}" "{parsed_dst}"'
        BaseRemoteCmdTask.__init__(
            self,
            name=name,
            remote_configs=remote_configs,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            executable=executable,
            pre_cmd=_RSYNC_UTIL_SCRIPT,
            cmd=cmd,
            cwd=cwd,
            upstreams=[ensure_rsync_is_installed] + upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            max_output_line=max_output_line,
            max_error_line=max_error_line,
            preexec_fn=preexec_fn,
            should_execute=should_execute,
        )

    def _get_parsed_path(self, is_remote: bool, path: str) -> str:
        if not is_remote:
            return path
        return "${_CONFIG_USER}@${_CONFIG_HOST}:" + path



from zrb.builtin.project.add.app._group import project_add_app_group
from zrb.builtin.project.add.app.generator import add_app_generator
from zrb.builtin.project.add.app.python import add_python_app

assert project_add_app_group
assert add_app_generator
assert add_python_app



from zrb.task_input.str_input import StrInput

module_name_input = StrInput(
    name="module-name",
    shortcut="m",
    description="module name",
    prompt="module name",
    default="library",
)



from zrb.builtin.base64._group import base64_group
from zrb.builtin.base64.decode import decode_base64
from zrb.builtin.base64.encode import encode_base64

assert base64_group
assert decode_base64
assert encode_base64



import os

from zrb.builtin.devtool.install._group import devtool_install_group
from zrb.builtin.devtool.install._helper import write_config
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask
from zrb.task.flow_task import FlowTask
from zrb.task.task import Task
from zrb.task_input.str_input import StrInput

_CURRENT_DIR = os.path.dirname(__file__)
_SHELL_SCRIPT_DIR = os.path.join(_CURRENT_DIR, "..", "..", "..", "..", "shell-scripts")


install_zsh = FlowTask(
    name="zsh",
    group=devtool_install_group,
    description="Zsh terminal + oh-my-zsh + zdharma",
    inputs=[
        StrInput(
            name="zsh-config-file",
            shortcut="c",
            prompt="Zsh config file",
            default="~/.zshrc",
        )
    ],
    steps=[
        CmdTask(
            name="install-zsh",
            cmd_path=[
                os.path.join(_SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(_CURRENT_DIR, "install.sh"),
            ],
            preexec_fn=None,
        ),
        Task(
            name="configure-zsh",
            run=write_config(
                template_file=os.path.join(_CURRENT_DIR, "resource", "config.sh"),
                config_file="{{input.zsh_config_file}}",
            ),
        ),
    ],
    retry=0,
)
runner.register(install_zsh)



import os
from typing import Any

from zrb.builtin.project._helper import validate_existing_project_dir
from zrb.builtin.project._input import project_dir_input
from zrb.builtin.project.add.app._group import project_add_app_group
from zrb.builtin.project.add.app.generator._helper import validate_existing_package
from zrb.builtin.project.add.app.generator._input import (
    generator_app_port_input,
    generator_base_image_input,
    generator_name_input,
    package_name_input,
)
from zrb.helper.accessories.color import colored
from zrb.helper.codemod.add_assert_resource import add_assert_resource
from zrb.helper.codemod.add_import_module import add_import_module
from zrb.helper.util import to_kebab_case, to_snake_case
from zrb.runner import runner
from zrb.task.decorator import python_task
from zrb.task.resource_maker import ResourceMaker
from zrb.task.task import Task

_CURRENT_DIR = os.path.dirname(__file__)


@python_task(
    name="validate",
    inputs=[project_dir_input, package_name_input],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get("project_dir")
    validate_existing_project_dir(project_dir)
    package_name = kwargs.get("package_name")
    validate_existing_package(project_dir, package_name)


copy_resource = ResourceMaker(
    name="copy-resource",
    inputs=[
        project_dir_input,
        package_name_input,
        generator_name_input,
        generator_base_image_input,
        generator_app_port_input,
    ],
    upstreams=[validate],
    replacements={
        "zrbPackageName": "{{input.package_name}}",
        "zrbGeneratorName": "{{input.generator_name}}",
        "zrbGeneratorBaseImage": "{{input.generator_base_image}}",
        "zrbGeneratorAppPort": "{{input.generator_app_port}}",
    },
    template_path=os.path.join(_CURRENT_DIR, "template"),
    destination_path="{{ input.project_dir }}",
    excludes=[
        "*/deployment/venv",
        "*/__pycache__",
    ],
)


@python_task(
    name="register-generator",
    inputs=[
        project_dir_input,
        package_name_input,
        generator_name_input,
    ],
    upstreams=[copy_resource],
)
def register_generator(*args: Any, **kwargs: Any):
    task = kwargs.get("_task")
    project_dir = kwargs.get("project_dir")
    package_name = kwargs.get("package_name")
    generator_name = kwargs.get("generator_name")
    kebab_package_name = to_kebab_case(package_name)
    snake_package_name = to_snake_case(package_name)
    snake_generator_name = to_snake_case(generator_name)
    package_init_path = os.path.join(
        project_dir, "src", kebab_package_name, "src", snake_package_name, "__init__.py"
    )
    with open(package_init_path, "r") as f:
        code = f.read()
    new_code = add_import_module(
        code=code,
        module_path=f"{snake_package_name}.{snake_generator_name}",
        alias=snake_generator_name,
    )
    new_code = add_assert_resource(code=new_code, resource=snake_generator_name)
    with open(package_init_path, "w") as f:
        f.write(new_code)
    task.print_out(colored("Register generator", color="yellow"))


@python_task(
    name="generator",
    group=project_add_app_group,
    description="Add app generator",
    upstreams=[
        register_generator,
    ],
    runner=runner,
)
async def add_app_generator(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    task.print_out(colored("Appp generator added", color="yellow"))



import os

from zrb.task_input.str_input import StrInput

SYSTEM_USER = os.getenv("USER", "incognito")

project_name_input = StrInput(
    name="project-name",
    shortcut="n",
    description="Project name",
    prompt="Project name (can be empty)",
    default="",
)

project_author_name_input = StrInput(
    name="project-author-name",
    prompt="Project author name",
    description="Project author name",
    default=SYSTEM_USER,
)

project_author_email_input = StrInput(
    name="project-author-email",
    prompt="Project author email",
    description="Project author email",
    default=f"{SYSTEM_USER}@gmail.com",
)

project_description_input = StrInput(
    name="project-description",
    description="Project description",
    prompt="Project description",
    default="Just another Zrb project",
)



from typing import Any

from zrb.builtin.git._group import git_group
from zrb.helper.accessories.color import colored
from zrb.helper.git.detect_changes import get_modified_file_states
from zrb.helper.task import show_lines
from zrb.runner import runner
from zrb.task.decorator import python_task
from zrb.task_input.bool_input import BoolInput
from zrb.task_input.str_input import StrInput


@python_task(
    name="get-file-changes",
    group=git_group,
    description="Get modified files",
    inputs=[
        StrInput(
            name="commit",
            shortcut="c",
            description="commit hash/tag",
            prompt="Commit hash/Tag",
            default="HEAD",
        ),
        BoolInput(
            name="include-new",
            description="include new files",
            prompt="Include new files",
            default=True,
        ),
        BoolInput(
            name="include-removed",
            description="include removed file",
            prompt="Include removed file",
            default=True,
        ),
        BoolInput(
            name="include-updated",
            description="include updated file",
            prompt="Include updated file",
            default=True,
        ),
    ],
    runner=runner,
)
async def get_git_file_changes(*args: Any, **kwargs: Any):
    commit = kwargs.get("commit", "HEAD")
    include_new = kwargs.get("include_new", True)
    include_removed = kwargs.get("include_removed", True)
    include_updated = kwargs.get("include_updated", True)
    modified_file_states = get_modified_file_states(commit)
    modified_file_keys = []
    output = []
    for modified_file, state in modified_file_states.items():
        if include_updated and state.minus and state.plus:
            output.append(colored(f"+- {modified_file}", color="yellow"))
            modified_file_keys.append(modified_file)
            continue
        if include_removed and state.minus and not state.plus:
            output.append(colored(f"-- {modified_file}", color="red"))
            modified_file_keys.append(modified_file)
            continue
        if include_new and state.plus and not state.minus:
            output.append(colored(f"++ {modified_file}", color="green"))
            modified_file_keys.append(modified_file)
            continue
    show_lines(kwargs["_task"], *output)
    modified_file_keys.sort()
    return "\n".join(modified_file_keys)



import autopep8

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.codemod.format_code", attrs=["dark"]))


@typechecked
def format_code(code: str) -> str:
    return autopep8.fix_code(code)



import os
from collections.abc import Mapping

from zrb.task.resource_maker import ResourceMaker


def copy_resource_replacement_mutator(
    task: ResourceMaker, replacements: Mapping[str, str]
) -> Mapping[str, str]:
    replacements["zrbBaseProjectDir"] = os.path.basename(
        replacements.get("zrbProjectDir", "")
    )
    if replacements.get("zrbProjectName", "") == "":
        replacements["zrbProjectName"] = replacements.get("zrbBaseProjectDir", "")
    return replacements



from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.task.base_task.base_task import BaseTask

logger.debug(colored("Loading zrb.task.task", attrs=["dark"]))


@typechecked
class Task(BaseTask):
    """
    Task is the smallest Zrb automation unit.

    You can configure a Task by using several interfaces:
    - `inputs`: interfaces to read user input at the beginning of the execution.
    - `envs`: interfaces to read and use OS Environment Variables.
    - `env_files`: interfaces to read and use Environment Files.

    Moreover, you can define Task dependencies by specifying its `upstreams` or by using shift-right operator.

    Every Zrb Task has its life-cycle state:
    - `Triggered`: The Task is triggered (either by the user or by the other Task).
    - `Waiting`: Zrb has already triggered the Task. The Task is now waiting for all its upstreams to be ready.
    - `Skipped`: Task upstreams are ready, but the Task is not executed and will immediately enter the `Ready` state.
    - `Started`: The upstreams are ready, and Zrb is now starting the Task execution.
    - `Failed`: Zrb failed to execute the Task. It will enter the `Retry` state if the current attempt does not exceed the maximum attempt.
    - `Retry`: The task has already entered the `Failed` state. Now, Zrb will try to start the Task execution.
    - `Ready`: The task is ready.

    There are several configurations related to Task's life cycle:
    - `retry`: Maximum retry attempt.
    - `retry_interval`: The duration is to wait before Zrb starts the next attempt.
    - `fallbacks`: Action to take if the Task has failed for good.
    - `checkers`: How to determine if a Task is `Ready`.
    - `checking_interval`: The duration to wait before Zrb checks for the Task's readiness.
    - `run`: Action to do when Zrb executes the Task.
    - `on_triggered`: Action to do when a Task is `Triggered`.
    - `on_waiting`: Action to do when a Task is `Waiting`.
    - `on_skipped`: Action to do when a Task is `Skipped`.
    - `on_started`: Action to do when a Task is `Started`.
    - `on_ready`: Action to do when a Task is `Ready`.
    - `on_retry`: Action to do when a Task is `Retry`.
    - `on_failed`: Action to do when a Task is `Failed`.
    - `should_execute`: Condition to determine whether a Task should be `Started` or `Skipped`.

    Finally, you can put related Tasks under the same `group`.
    """

    pass



import asyncio
import logging
from collections.abc import Callable
from typing import Any

from component.messagebus.messagebus import Admin, Consumer, Publisher
from component.rpc.rpc import Caller, Message, Result
from ulid import ULID


class MessagebusCaller(Caller):
    def __init__(
        self,
        logger: logging.Logger,
        admin: Admin,
        publisher: Publisher,
        consumer_factory: Callable[[], Consumer],
        timeout: float = 30,
        check_reply_interval: float = 0.05,
    ):
        self.logger = logger
        self.admin = admin
        self.publisher = publisher
        self.consumer_factory = consumer_factory
        self.timeout = timeout
        self.check_reply_interval = check_reply_interval

    async def call(self, rpc_name: str, *args: Any, **kwargs: Any):
        random_uuid = str(ULID())
        reply_event_name = f"reply_{rpc_name}_{random_uuid}"
        await self.admin.create_events([reply_event_name])
        reply_consumer = self.consumer_factory()
        reply_received_event = asyncio.Event()
        call_result, call_error = None, None

        @reply_consumer.register(reply_event_name)
        async def consume_reply(result_dict: Any):
            nonlocal call_result, call_error
            result = Result.from_dict(result_dict)
            call_result, call_error = result.result, result.error
            reply_received_event.set()
            await reply_consumer.stop()

        try:
            task = asyncio.create_task(reply_consumer.start())
            await self.publisher.publish(
                rpc_name,
                Message(
                    reply_event=reply_event_name, args=args, kwargs=kwargs
                ).to_dict(),
            )
            await asyncio.gather(task)
            await asyncio.wait_for(reply_received_event.wait(), timeout=self.timeout)
        except asyncio.TimeoutError:
            self.logger.error(
                "🤙 [messagebus-rpc-caller] "
                + f"Timeout while waiting for reply event {reply_event_name}"
            )
            raise Exception(
                f"Timeout while waiting for reply event: {reply_event_name}"
            )
        finally:
            await self._clean_up(reply_consumer, reply_event_name)
        # raise call_error or return call_result
        if call_error != "":
            self.logger.error(
                "🤙 [messagebus-rpc-caller] RPC "
                + f'"{rpc_name}" returning error: {call_error}'
            )
            raise Exception(call_error)
        self.logger.info(
            "🤙 [messagebus-rpc-caller] RPC "
            + f'"{rpc_name}" returning result: {call_result}'
        )
        return call_result

    async def _clean_up(self, reply_consumer: Consumer, reply_event_name: str):
        try:
            await reply_consumer.stop()
        except (asyncio.CancelledError, GeneratorExit, Exception):
            self.logger.error("🤙 [messagebus-rpc-caller]", exc_info=True)
        try:
            await self.admin.delete_events([reply_event_name])
        except (asyncio.CancelledError, GeneratorExit, Exception):
            self.logger.error("🤙 [messagebus-rpc-caller]", exc_info=True)



from typing import Any

from ..support._helper import get_support_container_compose_profiles


def activate_monolith_compose_profile(*args: Any, **kwargs: Any) -> str:
    compose_profiles = _get_monolith_container_compose_profiles(*args, **kwargs)
    compose_profile_str = ",".join(compose_profiles)
    return f"export COMPOSE_PROFILES={compose_profile_str}"


def should_start_monolith_container(*args: Any, **kwargs: Any) -> bool:
    if not kwargs.get("local_snake_zrb_app_name", True):
        return False
    compose_profiles = _get_monolith_container_compose_profiles(*args, **kwargs)
    return len(compose_profiles) > 0


def _get_monolith_container_compose_profiles(*args: Any, **kwargs: Any) -> list[str]:
    compose_profiles = get_support_container_compose_profiles(*args, **kwargs)
    compose_profiles.append("monolith")
    return compose_profiles



import logging

from config import APP_ENABLE_OTEL, APP_LOGGING_LEVEL, APP_OTEL_EXPORTER_OTLP_ENDPOINT
from opentelemetry._logs import set_logger_provider
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor

##############################################################################
# Create logger
##############################################################################
logger = logging.getLogger("src")
logger.setLevel(APP_LOGGING_LEVEL)

##############################################################################
# Formatter log handler
##############################################################################

stream_handler = logging.StreamHandler()
stream_handler.setLevel(APP_LOGGING_LEVEL)
# create formatter
formatter = logging.Formatter("%(levelname)s:\t%(message)s")
# set stream handler's formatter
stream_handler.setFormatter(formatter)
# add ch to logger
logger.addHandler(stream_handler)

##############################################################################
# Open telemetry log handler
##############################################################################

if APP_ENABLE_OTEL:
    # create logger providers
    otlp_logger_provider = LoggerProvider()
    # set the providers
    set_logger_provider(otlp_logger_provider)
    otlp_log_exporter = OTLPLogExporter(endpoint=APP_OTEL_EXPORTER_OTLP_ENDPOINT)
    # add the batch processors to the trace provider
    otlp_logger_provider.add_log_record_processor(
        BatchLogRecordProcessor(otlp_log_exporter)
    )
    otlp_log_handler = LoggingHandler(
        level=APP_LOGGING_LEVEL, logger_provider=otlp_logger_provider
    )
    logger.addHandler(otlp_log_handler)



from typing import Any

from zrb import Task, python_task, runner
from zrb.builtin import project_group
from zrb.helper.accessories.color import colored


@python_task(
    name="publish",
    group=project_group,
    description="Publish project artifacts",
    runner=runner,
)
def publish_project(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    task.print_out(colored("Project artifact published", color="yellow"))



from abc import ABC, abstractmethod


class Authorizer(ABC):
    @abstractmethod
    async def is_admin(self, user_id: str) -> bool:
        pass

    @abstractmethod
    async def is_guest(self, user_id: str) -> bool:
        pass

    @abstractmethod
    async def is_having_permission(self, user_id: str, *permission_names: str) -> bool:
        pass



from .append_all_disabled_env import append_all_disabled_env
from .append_all_enabled_env import append_all_enabled_env
from .append_deployment_template_env import append_deployment_template_env
from .append_src_template_env import append_src_template_env
from .create_app_config import create_app_config
from .create_microservice_config import create_microservice_config
from .register_migration import register_migration
from .register_module import register_module

assert append_all_disabled_env
assert append_all_enabled_env
assert append_deployment_template_env
assert append_src_template_env
assert create_app_config
assert create_microservice_config
assert register_migration
assert register_module



from zrb.builtin.git._group import git_group
from zrb.builtin.git.clear_branch import clear_git_branch
from zrb.builtin.git.get_file_changes import get_git_file_changes

assert git_group
assert clear_git_branch
assert get_git_file_changes



from zrb import Group
from zrb.builtin import project_group

snake_zrb_package_name_group = Group(
    name="snake_zrb_package_name",
    parent=project_group,
    description="Manage human readable zrb package name",
)



from config import APP_ENABLE_AUTH_MODULE
from helper.migration import migrate
from integration.db_connection import engine
from integration.log import logger
from module.auth.integration import Base
from module.auth.register_permission import register_permission


async def migrate_auth():
    if not APP_ENABLE_AUTH_MODULE:
        logger.info('🥪 Skip DB migration for "auth"')
        return
    logger.info('🥪 Perform DB migration for "auth"')
    await migrate(engine=engine, Base=Base)
    logger.info("🥪 Register permissions")
    await register_permission()



from zrb import DockerComposeTask, runner

from ..._project import remove_project_containers
from .._constant import RESOURCE_DIR
from ..image._env import image_env
from ._env import compose_env_file
from ._group import snake_zrb_app_name_container_group
from ._helper import activate_all_compose_profile
from ._service_config import snake_zrb_app_name_service_configs

remove_snake_zrb_app_name_container = DockerComposeTask(
    icon="💨",
    name="remove",
    description="Remove human readable zrb app name container",
    group=snake_zrb_app_name_container_group,
    cwd=RESOURCE_DIR,
    setup_cmd=activate_all_compose_profile,
    compose_cmd="down",
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
    compose_service_configs=snake_zrb_app_name_service_configs,
    env_files=[compose_env_file],
    envs=[image_env],
)

remove_snake_zrb_app_name_container >> remove_project_containers

runner.register(remove_snake_zrb_app_name_container)



import os

from zrb.builtin.monorepo._config import PROJECT_DIR
from zrb.task.cmd_task import CmdTask
from zrb.task_env.env import Env
from zrb.task_input.any_input import AnyInput
from zrb.task_input.str_input import StrInput

_CURRENT_DIR = os.path.dirname(__file__)


def create_pull_monorepo_task(
    task_name: str = "pull-monorepo",
) -> CmdTask:
    return CmdTask(
        name=task_name,
        envs=[
            Env("TIME", os_name="", default="{{datetime.datetime.now()}}"),
        ],
        cmd_path=[
            os.path.join(_CURRENT_DIR, "_common.sh"),
            os.path.join(_CURRENT_DIR, "pull-monorepo.sh"),
        ],
        cwd=PROJECT_DIR,
        should_show_cmd=False,
        should_show_working_directory=False,
        should_print_cmd_result=False,
        retry=0,
    )


def create_push_monorepo_task(
    task_name: str = "push-monorepo", default_commit_message: str = "Save changes"
) -> CmdTask:
    return CmdTask(
        name=task_name,
        inputs=[_create_commit_message_input(default_commit_message)],
        envs=[
            Env("MESSAGE", os_name="", default="{{input.message}}"),
        ],
        cmd_path=[
            os.path.join(_CURRENT_DIR, "_common.sh"),
            os.path.join(_CURRENT_DIR, "push-monorepo.sh"),
        ],
        cwd=PROJECT_DIR,
        should_show_cmd=False,
        should_show_working_directory=False,
        should_print_cmd_result=False,
        retry=0,
    )


def create_add_subrepo_task(
    origin: str,
    folder: str,
    branch: str,
    task_name: str = "add",
) -> CmdTask:
    return CmdTask(
        name=task_name,
        envs=[
            Env("ORIGIN", os_name="", default=origin),
            Env("FOLDER", os_name="", default=folder),
            Env("BRANCH", os_name="", default=branch),
            Env("TIME", os_name="", default="{{datetime.datetime.now()}}"),
        ],
        cmd_path=[
            os.path.join(_CURRENT_DIR, "_common.sh"),
            os.path.join(_CURRENT_DIR, "add-subrepo.sh"),
        ],
        cwd=PROJECT_DIR,
        should_show_cmd=False,
        should_show_working_directory=False,
        should_print_cmd_result=False,
        retry=0,
    )


def create_pull_subrepo_task(
    origin: str,
    folder: str,
    branch: str,
    task_name: str = "pull",
) -> CmdTask:
    return CmdTask(
        name=task_name,
        envs=[
            Env("ORIGIN", os_name="", default=origin),
            Env("FOLDER", os_name="", default=folder),
            Env("BRANCH", os_name="", default=branch),
            Env("TIME", os_name="", default="{{datetime.datetime.now()}}"),
        ],
        cmd_path=[
            os.path.join(_CURRENT_DIR, "_common.sh"),
            os.path.join(_CURRENT_DIR, "pull-subrepo.sh"),
        ],
        cwd=PROJECT_DIR,
        should_show_cmd=False,
        should_show_working_directory=False,
        should_print_cmd_result=False,
        retry=0,
    )


def create_push_subrepo_task(
    origin: str,
    folder: str,
    branch: str,
    task_name: str = "push",
    default_commit_message: str = "Commit",
) -> CmdTask:
    return CmdTask(
        name=task_name,
        inputs=[_create_commit_message_input(default_commit_message)],
        envs=[
            Env("ORIGIN", os_name="", default=origin),
            Env("FOLDER", os_name="", default=folder),
            Env("BRANCH", os_name="", default=branch),
            Env("TIME", os_name="", default="{{datetime.datetime.now()}}"),
            Env("MESSAGE", os_name="", default="{{input.message}}"),
        ],
        cmd_path=[
            os.path.join(_CURRENT_DIR, "_common.sh"),
            os.path.join(_CURRENT_DIR, "push-subrepo.sh"),
        ],
        cwd=PROJECT_DIR,
        should_show_cmd=False,
        should_show_working_directory=False,
        should_print_cmd_result=False,
        retry=0,
    )


def _create_commit_message_input(default: str) -> AnyInput:
    return StrInput(
        name="message",
        shortcut="m",
        prompt="Commit Message",
        default=default,
    )



import os
import pathlib
from collections.abc import Callable, Iterable, Mapping
from typing import Any, Optional, Union

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import JinjaTemplate
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.cmd_task import CmdVal
from zrb.task.docker_compose_task import DockerComposeTask, ServiceConfig
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.docker_compose_task", attrs=["dark"]))


@typechecked
class DockerComposeStartTask(DockerComposeTask):
    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        executable: Optional[str] = None,
        compose_service_configs: Mapping[str, ServiceConfig] = {},
        compose_file: Optional[str] = None,
        compose_options: Mapping[JinjaTemplate, JinjaTemplate] = {},
        compose_flags: Iterable[JinjaTemplate] = [],
        compose_args: Iterable[JinjaTemplate] = [],
        compose_env_prefix: str = "",
        setup_cmd: CmdVal = "",
        setup_cmd_path: CmdVal = "",
        cwd: Optional[Union[JinjaTemplate, pathlib.Path]] = None,
        should_render_cwd: bool = True,
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0.05,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        preexec_fn: Optional[Callable[[], Any]] = os.setsid,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
        return_upstream_result: bool = False,
        should_print_cmd_result: bool = True,
        should_show_cmd: bool = True,
        should_show_working_directory: bool = True,
    ):
        DockerComposeTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            executable=executable,
            compose_service_configs=compose_service_configs,
            compose_file=compose_file,
            compose_options=compose_options,
            compose_flags=compose_flags,
            compose_args=compose_args,
            compose_env_prefix=compose_env_prefix,
            setup_cmd=setup_cmd,
            setup_cmd_path=setup_cmd_path,
            cwd=cwd,
            should_render_cwd=should_render_cwd,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            max_output_line=max_output_line,
            max_error_line=max_error_line,
            preexec_fn=preexec_fn,
            should_execute=should_execute,
            return_upstream_result=return_upstream_result,
            should_print_cmd_result=should_print_cmd_result,
            should_show_cmd=should_show_cmd,
            should_show_working_directory=should_show_working_directory,
        )

    def get_cmd_script(self, *args: Any, **kwargs: Any) -> str:
        # setup
        setup_cmd = self._create_cmd_script(
            self._setup_cmd_path, self._setup_cmd, *args, **kwargs
        )
        cmd_list = [setup_cmd] if setup_cmd.strip() != "" else []
        # compose command
        cmd_list = cmd_list + [
            # compose start
            self._get_docker_compose_cmd_script(
                compose_cmd="up",
                compose_options=self._compose_options,
                compose_flags=list(self._compose_flags) + ["-d"],
                compose_args=self._compose_args,
                *args,
            ),
            # compose log
            self._get_docker_compose_cmd_script(
                compose_cmd="logs",
                compose_options={},
                compose_flags=["-f"],
                compose_args=[],
                *args,
            ),
        ]
        cmd_str = "\n".join(cmd_list)
        self.log_info(f"Command: {cmd_str}")
        return cmd_str



import asyncio
import copy
from collections.abc import Callable, Iterable, Mapping
from typing import Any, Optional, Union

from zrb.helper.accessories.color import colored
from zrb.helper.accessories.name import get_random_name
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_kebab_case
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.base_task.base_task import BaseTask
from zrb.task.flow_task import FlowTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.server", attrs=["dark"]))


@typechecked
class Controller:
    def __init__(
        self,
        trigger: Union[AnyTask, list[AnyTask]],
        action: Union[AnyTask, list[AnyTask]],
        name: Optional[str] = None,
    ):
        self._name = get_random_name() if name is None else name
        self._triggers = self._to_task_list(trigger)
        self._actions = self._to_task_list(action)
        self._args: list[Any] = []
        self._kwargs: Mapping[str, Any] = {}
        self._inputs: list[AnyInput] = []
        self._envs: list[Env] = []
        self._env_files: list[EnvFile] = []

    def set_args(self, args: list[Any]):
        self._args = args

    def set_kwargs(self, kwargs: Mapping[str, Any]):
        self._kwargs = kwargs

    def set_inputs(self, inputs: list[AnyInput]):
        self._inputs = inputs

    def set_envs(self, envs: list[Env]):
        self._envs = envs

    def set_env_files(self, env_files: list[EnvFile]):
        self._env_files = env_files

    def get_original_env_files(self) -> Iterable[EnvFile]:
        env_files = []
        for trigger in self._triggers:
            env_files += trigger._get_env_files()
        for action in self._actions:
            env_files += action._get_env_files()
        return env_files

    def get_original_envs(self) -> Iterable[Env]:
        envs = []
        for trigger in self._triggers:
            envs += trigger._get_envs()
        for action in self._actions:
            envs += action._get_envs()
        return envs

    def get_original_inputs(self) -> Iterable[AnyInput]:
        inputs = []
        for trigger in self._triggers:
            inputs += trigger._get_combined_inputs()
        for action in self._actions:
            inputs += action._get_combined_inputs()
        return inputs

    def to_function(self) -> Callable[..., Any]:
        task = self._get_task()

        async def fn() -> Any:
            task.print_out_dark(f"Starting controller: {self._name}")
            task_fn = task.to_function(is_async=True)
            return await task_fn(*self._args, **self._kwargs)

        return fn

    def _to_task_list(self, tasks: Union[AnyTask, list[AnyTask]]) -> list[AnyTask]:
        if isinstance(tasks, AnyTask):
            return [tasks.copy()]
        return [task.copy() for task in tasks]

    def _get_task(self) -> AnyTask:
        actions = [action.copy() for action in self._actions]
        actions.insert(0, self._get_remonitor_task())
        triggers = [trigger.copy() for trigger in self._triggers]
        task: AnyTask = FlowTask(
            name=to_kebab_case(self._name),
            inputs=self._inputs,
            envs=self._envs,
            env_files=self._env_files,
            steps=[triggers, actions],
        )
        return task

    def _get_remonitor_task(self) -> AnyTask:
        async def on_ready(task: AnyTask):
            task = self._get_task()
            fn = task.to_function(is_async=True)
            await fn()

        return BaseTask(
            name=f"monitor-{to_kebab_case(self._name)}",
            on_ready=on_ready,
        )


@typechecked
class Server(BaseTask):

    def __init__(
        self,
        name: str,
        controllers: Iterable[Controller],
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[int, float] = 0.05,
        retry: int = 0,
        retry_interval: Union[int, float] = 1,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
        return_upstream_result: bool = False,
    ):
        inputs, envs, env_files = list(inputs), list(envs), list(env_files)
        for controller in controllers:
            controller_cp = copy.deepcopy(controller)
            inputs += controller_cp.get_original_inputs()
            envs += controller_cp.get_original_envs()
            env_files += controller_cp.get_original_env_files()
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            should_execute=should_execute,
            return_upstream_result=return_upstream_result,
        )
        self._controllers = list(controllers)

    async def run(self, *args: Any, **kwargs: Any):
        for controller in self._controllers:
            controller.set_envs(self._get_envs())
            controller.set_env_files(self._get_env_files())
            controller.set_inputs(self._get_inputs())
            controller.set_args(args)
            controller.set_kwargs(kwargs)
        functions = [controller.to_function() for controller in self._controllers]
        await asyncio.gather(*[fn() for fn in functions])



from zrb.builtin.devtool.install.aws.aws import install_aws
from zrb.builtin.devtool.install.docker.docker import install_docker
from zrb.builtin.devtool.install.gcloud.gcloud import install_gcloud
from zrb.builtin.devtool.install.gvm.gvm import install_gvm
from zrb.builtin.devtool.install.helix.helix import install_helix
from zrb.builtin.devtool.install.helm.helm import install_helm
from zrb.builtin.devtool.install.kubectl.kubectl import install_kubectl
from zrb.builtin.devtool.install.nvm.nvm import install_nvm
from zrb.builtin.devtool.install.pulumi.pulumi import install_pulumi
from zrb.builtin.devtool.install.pyenv.pyenv import install_pyenv
from zrb.builtin.devtool.install.sdkman.sdkman import install_sdkman
from zrb.builtin.devtool.install.selenium.selenium import install_selenium
from zrb.builtin.devtool.install.terraform.terraform import install_terraform
from zrb.builtin.devtool.install.tmux.tmux import install_tmux
from zrb.builtin.devtool.install.zsh.zsh import install_zsh
from zrb.builtin.ubuntu.install._group import ubuntu_install_group
from zrb.builtin.ubuntu.install.essentials import install_ubuntu_essentials
from zrb.builtin.ubuntu.install.tex import install_ubuntu_tex
from zrb.builtin.ubuntu.install.toys import install_ubuntu_toys
from zrb.builtin.ubuntu.update import update_ubuntu
from zrb.runner import runner
from zrb.task.flow_task import FlowTask

install_ubuntu_all = FlowTask(
    name="all",
    group=ubuntu_install_group,
    description="Install all ubuntu packages",
    steps=[
        update_ubuntu,
        install_ubuntu_essentials,
        install_ubuntu_toys,
        install_ubuntu_tex,
        install_zsh,
        install_tmux,
        install_gvm,
        install_nvm,
        install_pyenv,
        install_sdkman,
        install_aws,
        install_gcloud,
        install_docker,
        install_kubectl,
        install_terraform,
        install_helm,
        install_pulumi,
        install_helix,
        install_selenium,
    ],
)
runner.register(install_ubuntu_all)



import os

from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_kebab_case, to_snake_case


@typechecked
def validate_existing_package(project_dir: str, package_name: str):
    kebab_package_name = to_kebab_case(package_name)
    snake_package_name = to_snake_case(package_name)
    package_init_path = os.path.join(
        project_dir, "src", kebab_package_name, "src", snake_package_name, "__init__.py"
    )
    if not os.path.isfile(package_init_path):
        raise Exception(f"Package not exists: {package_init_path}")



import datetime
import random
import re
import sys
from collections.abc import Iterable
from typing import Optional

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.advertisement", attrs=["dark"]))


@typechecked
class Advertisement:
    def __init__(self, content: str, time_pattern: str = ".*"):
        self.time_pattern = time_pattern
        self.content = content

    def show(self):
        print(colored(self.content, attrs=["dark"]), file=sys.stderr)


@typechecked
def get_advertisement(
    advertisements: Iterable[Advertisement],
) -> Optional[Advertisement]:
    now = datetime.datetime.now().isoformat()
    candidates = [adv for adv in advertisements if re.match(adv.time_pattern, now)]
    if len(candidates) == 0:
        return None
    return random.choice(candidates)



import random
import string

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.accessories.name", attrs=["dark"]))


@typechecked
def get_random_name(
    separator: str = "-", add_random_digit: bool = True, digit_count: int = 4
) -> str:
    prefixes = [
        "bold",
        "calm",
        "dark",
        "deep",
        "fast",
        "firm",
        "glad",
        "grey",
        "hard",
        "high",
        "kind",
        "late",
        "lean",
        "long",
        "loud",
        "mild",
        "neat",
        "pure",
        "rare",
        "rich",
        "safe",
        "slow",
        "soft",
        "tall",
        "thin",
        "trim",
        "vast",
        "warm",
        "weak",
        "wild",
    ]
    suffixes = [
        "arch",
        "area",
        "atom",
        "base",
        "beam",
        "bell",
        "bolt",
        "bone",
        "bulk",
        "bush",
        "cell",
        "chip",
        "clay",
        "coal",
        "coil",
        "cone",
        "cube",
        "disk",
        "dust",
        "face",
        "film",
        "foam",
        "frog",
        "fuel",
        "gate",
        "gear",
        "hall",
        "hand",
        "horn",
        "leaf",
    ]
    prefix = random.choice(prefixes)
    suffix = random.choice(suffixes)
    parts = [prefix, suffix]
    if add_random_digit:
        random_digit = "".join(random.choices(string.digits, k=digit_count))
        parts.append(random_digit)
    return separator.join(parts)



from datetime import datetime

from zrb.builtin.monorepo._group import monorepo_group
from zrb.builtin.monorepo._task import PUSH_SUBREPO_UPSTREAM
from zrb.runner import runner
from zrb.task.task import Task
from zrb.task_input.str_input import StrInput

push_to_monorepo = Task(
    name="push",
    group=monorepo_group,
    description="Pushing to subrepos",
    inputs=[
        StrInput(
            name="message",
            shortcut="m",
            prompt="Commit Messsage",
            default=lambda m: f"Synchronize subrepos at {datetime.now()}",
        )
    ],
    retry=0,
)
PUSH_SUBREPO_UPSTREAM >> push_to_monorepo
runner.register(push_to_monorepo)



from typing import Any

from zrb import Task, python_task, runner
from zrb.helper.accessories.color import colored

from ._group import project_container_group


@python_task(
    name="start",
    group=project_container_group,
    description="Start project containers",
    runner=runner,
)
def start_project_containers(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    task.print_out(colored("Project containers started", color="yellow"))



from zrb import DockerComposeStartTask, HTTPChecker, runner

from ..._project import start_project_containers
from .._constant import RESOURCE_DIR
from .._input import host_input, https_input, local_input
from ..image import build_snake_zrb_app_name_image
from ..image._env import image_env
from ..image._input import image_input
from ._env import compose_env_file, host_port_env
from ._group import snake_zrb_app_name_container_group
from ._service_config import snake_zrb_app_name_service_config
from .remove import remove_snake_zrb_app_name_container

start_snake_zrb_app_name_container = DockerComposeStartTask(
    icon="🐳",
    name="start",
    description="Start human readable zrb app name container",
    group=snake_zrb_app_name_container_group,
    inputs=[
        local_input,
        host_input,
        https_input,
        image_input,
    ],
    should_execute="{{ input.local_snake_zrb_app_name}}",
    upstreams=[build_snake_zrb_app_name_image, remove_snake_zrb_app_name_container],
    cwd=RESOURCE_DIR,
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
    compose_service_configs={"kebab-zrb-app-name": snake_zrb_app_name_service_config},
    env_files=[compose_env_file],
    envs=[
        image_env,
        host_port_env,
    ],
    checkers=[
        HTTPChecker(
            name="check-kebab-zrb-app-name",
            host="{{input.snake_zrb_app_name_host}}",
            port="{{env.HOST_PORT}}",
            is_https="{{input.snake_zrb_app_name_https}}",
        )
    ],
)

start_snake_zrb_app_name_container >> start_project_containers

runner.register(start_snake_zrb_app_name_container)



from zrb.builtin.process._group import process_group
from zrb.builtin.process.pid import (
    get_process_pid_by_name,
    get_process_pid_by_port,
    process_pid_group,
)

assert process_group
assert process_pid_group
assert get_process_pid_by_name
assert get_process_pid_by_port



import asyncio
import os
from typing import Any

from zrb.builtin.project._helper import validate_existing_project_dir
from zrb.builtin.project._input import project_dir_input
from zrb.builtin.project.add.fastapp._group import project_add_fastapp_group
from zrb.builtin.project.add.fastapp.app._input import app_name_input
from zrb.builtin.project.add.fastapp.module._helper import (
    append_all_disabled_env,
    append_all_enabled_env,
    append_deployment_template_env,
    append_src_template_env,
    create_app_config,
    create_microservice_config,
    register_migration,
    register_module,
)
from zrb.builtin.project.add.fastapp.module._input import module_name_input
from zrb.helper import util
from zrb.helper.accessories.color import colored
from zrb.runner import runner
from zrb.task.decorator import python_task
from zrb.task.resource_maker import ResourceMaker
from zrb.task.task import Task

_CURRENT_DIR = os.path.dirname(__file__)


@python_task(
    name="validate",
    inputs=[project_dir_input, app_name_input, module_name_input],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get("project_dir")
    validate_existing_project_dir(project_dir)
    app_name = kwargs.get("app_name")
    module_name = kwargs.get("module_name")
    snake_module_name = util.to_snake_case(module_name)
    automation_dir = os.path.join(
        project_dir, "_automate", util.to_snake_case(app_name)
    )
    if not os.path.exists(automation_dir):
        raise Exception(f"Automation directory does not exist: {automation_dir}")
    app_dir = os.path.join(project_dir, "src", f"{util.to_kebab_case(app_name)}")
    if not os.path.exists(app_dir):
        raise Exception(f"App directory does not exist: {app_dir}")
    module_path = os.path.join(app_dir, "src", "module", snake_module_name)
    if os.path.exists(module_path):
        raise Exception(f"Module directory already exists: {module_path}")


copy_resource = ResourceMaker(
    name="copy-resource",
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
    ],
    upstreams=[validate],
    replacements={
        "zrbAppName": "{{input.app_name}}",
        "zrbModuleName": "{{input.module_name}}",
    },
    template_path=os.path.join(_CURRENT_DIR, "template"),
    destination_path="{{ input.project_dir }}",
    excludes=[
        "*/__pycache__",
    ],
)


@python_task(
    name="module",
    group=project_add_fastapp_group,
    description="Add Fastapp module",
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
    ],
    upstreams=[copy_resource],
    runner=runner,
)
async def add_fastapp_module(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    project_dir = kwargs.get("project_dir", ".")
    app_name = kwargs.get("app_name")
    module_name = kwargs.get("module_name")
    await asyncio.gather(
        asyncio.create_task(
            create_microservice_config(task, project_dir, app_name, module_name)
        ),
        asyncio.create_task(register_module(task, project_dir, app_name, module_name)),
        asyncio.create_task(
            register_migration(task, project_dir, app_name, module_name)
        ),
        asyncio.create_task(
            create_app_config(task, project_dir, app_name, module_name)
        ),
        asyncio.create_task(
            append_all_enabled_env(task, project_dir, app_name, module_name)
        ),
        asyncio.create_task(
            append_all_disabled_env(task, project_dir, app_name, module_name)
        ),
        asyncio.create_task(
            append_src_template_env(task, project_dir, app_name, module_name)
        ),
        asyncio.create_task(
            append_deployment_template_env(task, project_dir, app_name, module_name)
        ),
    )
    task.print_out(colored("Fastapp crud added", color="yellow"))



from typing import Any

from zrb import Task


def activate_support_compose_profile(*args: Any, **kwargs: Any) -> str:
    compose_profiles = get_support_container_compose_profiles(*args, **kwargs)
    compose_profile_str = ",".join(compose_profiles)
    return f"export COMPOSE_PROFILES={compose_profile_str}"


def should_start_support_container(*args: Any, **kwargs: Any) -> bool:
    if not kwargs.get("local_snake_zrb_app_name", True):
        return False
    compose_profiles = get_support_container_compose_profiles(*args, **kwargs)
    return len(compose_profiles) > 0


def get_support_container_compose_profiles(*args: Any, **kwargs: Any) -> list[str]:
    task: Task = kwargs.get("_task")
    env_map = task.get_env_map()
    compose_profiles: list[str] = []
    if env_map.get("APP_DB_CONNECTION", "").startswith("postgresql"):
        compose_profiles.append("postgres")
    broker_type = env_map.get("APP_BROKER_TYPE", "rabbitmq")
    if broker_type in ["rabbitmq", "kafka"]:
        compose_profiles.append(broker_type)
    if kwargs.get("enable_snake_zrb_app_name_monitoring", False):
        compose_profiles.append("monitoring")
    return compose_profiles



import os

from zrb import CmdTask, Env, runner

from ..._project import deploy_project
from .._constant import AUTOMATE_DIR, DEPLOYMENT_DIR, PREFER_MICROSERVICES
from .._env import (
    deployment_app_env_file,
    deployment_config_env_file,
    deployment_enable_monitoring_env,
    deployment_modules_env,
    pulumi_backend_url_env,
    pulumi_config_passphrase_env,
)
from .._input import pulumi_stack_input, replica_input
from ..container._input import enable_monitoring_input
from ..image import push_snake_zrb_app_name_image
from ..image._env import image_env
from ..image._input import image_input
from ._group import snake_zrb_app_name_microservices_group

_CURRENT_DIR = os.path.dirname(__file__)

deploy_snake_zrb_app_name_microservices = CmdTask(
    icon="🚧",
    name="deploy",
    description="Deploy human readable zrb app name as a microservices",
    group=snake_zrb_app_name_microservices_group,
    inputs=[
        image_input,
        replica_input,
        pulumi_stack_input,
        enable_monitoring_input,
    ],
    upstreams=[push_snake_zrb_app_name_image],
    cwd=DEPLOYMENT_DIR,
    env_files=[
        deployment_config_env_file,
        deployment_app_env_file,
    ],
    envs=[
        pulumi_backend_url_env,
        pulumi_config_passphrase_env,
        image_env,
        deployment_modules_env,
        deployment_enable_monitoring_env,
        Env(
            name="MODE",
            os_name="DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX_MODE",
            default="microservices",
        ),
    ],
    cmd_path=[
        os.path.join(AUTOMATE_DIR, "init-pulumi-stack.sh"),
        os.path.join(_CURRENT_DIR, "deploy.sh"),
    ],
)

if PREFER_MICROSERVICES:
    deploy_snake_zrb_app_name_microservices >> deploy_project

runner.register(deploy_snake_zrb_app_name_microservices)



import base64
from typing import Any

from zrb.builtin.base64._group import base64_group
from zrb.builtin.base64._input import text_input
from zrb.runner import runner
from zrb.task.decorator import python_task


@python_task(
    name="encode",
    group=base64_group,
    inputs=[text_input],
    description="Encode a text using base64 algorithm",
    retry=0,
    runner=runner,
)
async def encode_base64(*args: str, **kwargs: Any):
    text: str = kwargs.get("text", "")
    encoded_text = base64.b64encode(text.encode())
    return encoded_text.decode()



from config import APP_AUTH_ACCESS_TOKEN_COOKIE_KEY
from module.auth.component import (
    AccessTokenScheme,
    create_oauth2_bearer_access_token_scheme,
)
from module.auth.integration.access_token_util import access_token_util
from module.auth.integration.user import guest_user

access_token_scheme: AccessTokenScheme = (
    create_oauth2_bearer_access_token_scheme(  # noqa
        guest_user=guest_user,
        access_token_util=access_token_util,
        token_url="/api/v1/auth/login-oauth",
        token_cookie_key=APP_AUTH_ACCESS_TOKEN_COOKIE_KEY,
    )
)



from typing import Type, TypeVar

import jsons
from component.messagebus.messagebus import Publisher
from component.model.repo_model import RepoModel
from component.repo.repo import Repo
from module.log.schema.activity import ActivityData
from pydantic import BaseModel

Schema = TypeVar("Schema", bound=BaseModel)
SchemaData = TypeVar("SchemaData", bound=BaseModel)
SchemaResult = TypeVar("SchemaResult", bound=BaseModel)


class HistoricalRepoModel(RepoModel[Schema, SchemaData, SchemaResult]):
    schema_result_cls: Type[SchemaResult]
    log_entity_name: str

    def __init__(self, repo: Repo[Schema, SchemaData], publisher: Publisher):
        super().__init__(repo)
        self.publisher = publisher

    async def insert(self, data: SchemaData) -> Schema:
        result = await super().insert(data)
        await self._publish_activity(action="insert", result=result)
        return result

    async def update(self, id: str, data: SchemaData) -> Schema:
        result = await self.repo.update(id, data)
        await self._publish_activity(action="update", result=result)
        return result

    async def delete(self, id: str) -> Schema:
        result = await self.repo.delete(id)
        await self._publish_activity(action="delete", result=result)
        return result

    async def _publish_activity(self, action: str, result: Schema):
        activity_data = ActivityData(
            action=action,
            entity=self.log_entity_name,
            data=jsons.dumps(result.model_dump()),
        )
        await self.publisher.publish("log_new_activity", activity_data.model_dump())



from collections.abc import AsyncIterator

import pytest
from httpx import AsyncClient
from src.config import APP_NAME


@pytest.mark.asyncio
async def test_get_liveness(test_client_generator: AsyncIterator[AsyncClient]):
    async for client in test_client_generator:
        response = await client.get("/liveness")
        assert response.status_code == 200
        assert response.json() == {"app": APP_NAME, "alive": True}


@pytest.mark.asyncio
async def test_head_liveness(test_client_generator: AsyncIterator[AsyncClient]):
    async for client in test_client_generator:
        response = await client.head("/liveness")
        assert response.status_code == 200


@pytest.mark.asyncio
async def test_get_readiness(test_client_generator: AsyncIterator[AsyncClient]):
    async for client in test_client_generator:
        response = await client.get("/readiness")
        assert response.status_code == 200
        assert response.json() == {"app": APP_NAME, "ready": True}


@pytest.mark.asyncio
async def test_head_readiness(test_client_generator: AsyncIterator[AsyncClient]):
    async for client in test_client_generator:
        response = await client.head("/readiness")
        assert response.status_code == 200



import random
from collections.abc import Iterable
from typing import Optional

from termcolor import COLORS

from zrb.helper.accessories.untyped_color import untyped_colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(untyped_colored("Loading zrb.helper.accessories.color", attrs=["dark"]))


@typechecked
def get_random_color() -> str:
    colors = [
        "green",
        "yellow",
        "blue",
        "magenta",
        "cyan",
        "white",
        "light_green",
        "light_yellow",
        "light_blue",
        "light_magenta",
        "light_cyan",
    ]
    return random.choice(colors)


@typechecked
def is_valid_color(color: str) -> bool:
    return color in COLORS


@typechecked
def colored(
    text: str,
    color: Optional[str] = None,
    on_color: Optional[str] = None,
    attrs: Optional[Iterable[str]] = None,
) -> str:
    return untyped_colored(text, color, on_color, attrs)



import os

from zrb.helper.codemod.add_import_module import add_import_module
from zrb.helper.codemod.append_code_to_function import append_code_to_function
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_snake_case
from zrb.task.task import Task

from ._common import get_app_module_dir


@typechecked
async def register_rpc(
    task: Task, project_dir: str, app_name: str, module_name: str, entity_name: str
):
    snake_module_name = to_snake_case(module_name)
    snake_entity_name = to_snake_case(entity_name)
    module_rpc_file_path = os.path.join(
        get_app_module_dir(project_dir, app_name), snake_module_name, "rpc.py"
    )
    register_function_path = ".".join(
        ["module", snake_module_name, "entity", snake_entity_name, "rpc"]
    )
    register_function = f"register_{snake_entity_name}_rpc"
    task.print_out(f"Read code from: {module_rpc_file_path}")
    code = await read_text_file_async(module_rpc_file_path)
    task.print_out(
        f'Add import "register_rpc" as "{register_function}" '
        + f'from "{register_function_path}" to the code'
    )
    code = add_import_module(
        code=code,
        module_path=register_function_path,
        resource="register_rpc",
        alias=register_function,
    )
    task.print_out(f'Add "{register_function}" call to the code')
    code = append_code_to_function(
        code=code,
        function_name="register_rpc",
        new_code=f"{register_function}(logger, rpc_server, rpc_caller, publisher)",  # noqa
    )
    task.print_out(f"Write modified code to: {module_rpc_file_path}")
    await write_text_file_async(module_rpc_file_path, code)



from collections.abc import AsyncIterator
from datetime import date, datetime, time

import pytest
from config import APP_AUTH_ADMIN_PASSWORD, APP_AUTH_ADMIN_USERNAME
from httpx import AsyncClient

inserted_success_data = {
    "snake_zrb_column_name": "test-kebab-create-entity-name-success"
}
to_be_updated_success_data = {
    "snake_zrb_column_name": "test-kebab-zrb-entity-name-to-be-updated-success"
}
updated_success_data = {
    "snake_zrb_column_name": "test-kebab-zrb-entity-name-updated-success"
}
to_be_deleted_success_data = {
    "snake_zrb_column_name": "test-kebab-zrb-entity-name-to-be-deleted-success"
}


@pytest.mark.asyncio
async def test_insert_snake_zrb_entity_name_and_get_success(
    test_client_generator: AsyncIterator[AsyncClient],
):
    async for client in test_client_generator:
        # login
        login_admin_response = await client.post(
            "/api/v1/auth/login",
            json={
                "identity": APP_AUTH_ADMIN_USERNAME,
                "password": APP_AUTH_ADMIN_PASSWORD,
            },
        )
        assert login_admin_response.status_code == 200
        admin_access_token = login_admin_response.json().get("access_token", "")

        # create snake_zrb_entity_name
        create_response = await client.post(
            "/api/v1/snake_zrb_module_name/kebab-zrb-plural-entity-name",
            json=inserted_success_data,
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert create_response.status_code == 200
        json_create_response = create_response.json()
        create_response_id = json_create_response.get("id", "")
        assert create_response_id != ""

        # get_by_id
        get_by_id_response = await client.get(
            f"/api/v1/snake_zrb_module_name/kebab-zrb-plural-entity-name/{create_response_id}",
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert get_by_id_response.status_code == 200
        json_get_by_id_response = get_by_id_response.json()
        get_by_id_response_id = json_get_by_id_response.get("id", "")
        assert get_by_id_response_id == create_response_id

        # get
        get_response = await client.get(
            "/api/v1/snake_zrb_module_name/kebab-zrb-plural-entity-name",
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert get_response.status_code == 200
        json_get_response = get_response.json()
        get_response_count = json_get_response.get("count", "")
        assert get_response_count > 0
        get_response_data = json_get_response.get("data", [])
        get_created_response_data = [
            row for row in get_response_data if row["id"] == create_response_id
        ]
        assert len(get_created_response_data) == 1


@pytest.mark.asyncio
async def test_update_snake_zrb_entity_name_and_get_success(
    test_client_generator: AsyncIterator[AsyncClient],
):
    async for client in test_client_generator:
        # login
        login_admin_response = await client.post(
            "/api/v1/auth/login",
            json={
                "identity": APP_AUTH_ADMIN_USERNAME,
                "password": APP_AUTH_ADMIN_PASSWORD,
            },
        )
        assert login_admin_response.status_code == 200
        admin_access_token = login_admin_response.json().get("access_token", "")

        # create snake_zrb_entity_name
        create_response = await client.post(
            "/api/v1/snake_zrb_module_name/kebab-zrb-plural-entity-name",
            json=to_be_updated_success_data,
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert create_response.status_code == 200
        json_create_response = create_response.json()
        create_response_id = json_create_response.get("id", "")
        assert create_response_id != ""

        # update snake_zrb_entity_name
        update_response = await client.put(
            f"/api/v1/snake_zrb_module_name/kebab-zrb-plural-entity-name/{create_response_id}",
            json=updated_success_data,
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert update_response.status_code == 200
        json_update_response = update_response.json()
        update_response_id = json_update_response.get("id", "")
        assert update_response_id == create_response_id

        # get_by_id
        get_by_id_response = await client.get(
            f"/api/v1/snake_zrb_module_name/kebab-zrb-plural-entity-name/{create_response_id}",
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert get_by_id_response.status_code == 200
        json_get_by_id_response = get_by_id_response.json()
        get_by_id_response_id = json_get_by_id_response.get("id", "")
        assert get_by_id_response_id == create_response_id
        for key, expected_value in updated_success_data.items():
            actual_value = json_get_by_id_response.get(key, "")
            assert f"{key}:{actual_value}" == f"{key}:{expected_value}"

        # get
        get_response = await client.get(
            "/api/v1/snake_zrb_module_name/kebab-zrb-plural-entity-name",
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert get_response.status_code == 200
        json_get_response = get_response.json()
        get_response_count = json_get_response.get("count", "")
        assert get_response_count > 0
        get_response_data = json_get_response.get("data", [])
        get_created_response_data = [
            row for row in get_response_data if row["id"] == create_response_id
        ]
        assert len(get_created_response_data) == 1


@pytest.mark.asyncio
async def test_delete_snake_zrb_entity_name_and_get_success(
    test_client_generator: AsyncIterator[AsyncClient],
):
    async for client in test_client_generator:
        # login
        login_admin_response = await client.post(
            "/api/v1/auth/login",
            json={
                "identity": APP_AUTH_ADMIN_USERNAME,
                "password": APP_AUTH_ADMIN_PASSWORD,
            },
        )
        assert login_admin_response.status_code == 200
        admin_access_token = login_admin_response.json().get("access_token", "")

        # create snake_zrb_entity_name
        create_response = await client.post(
            "/api/v1/snake_zrb_module_name/kebab-zrb-plural-entity-name",
            json=to_be_deleted_success_data,
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert create_response.status_code == 200
        json_create_response = create_response.json()
        create_response_id = json_create_response.get("id", "")
        assert create_response_id != ""

        # create snake_zrb_entity_name
        delete_response = await client.delete(
            f"/api/v1/snake_zrb_module_name/kebab-zrb-plural-entity-name/{create_response_id}",
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert delete_response.status_code == 200
        json_delete_response = delete_response.json()
        delete_response_id = json_delete_response.get("id", "")
        assert delete_response_id == create_response_id

        # get_by_id
        get_by_id_response = await client.get(
            f"/api/v1/snake_zrb_module_name/kebab-zrb-plural-entity-name/{create_response_id}",
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert get_by_id_response.status_code == 404

        # get
        get_response = await client.get(
            "/api/v1/snake_zrb_module_name/kebab-zrb-plural-entity-name",
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert get_response.status_code == 200
        json_get_response = get_response.json()
        get_response_count = json_get_response.get("count", "")
        assert get_response_count > 0
        get_response_data = json_get_response.get("data", [])
        get_created_response_data = [
            row for row in get_response_data if row["id"] == create_response_id
        ]
        assert len(get_created_response_data) == 0



from module.auth.integration import Base
from sqlalchemy import Column, ForeignKey, String, Table

user_group = Table(
    "user_groups",
    Base.metadata,
    Column("user_id", String, ForeignKey("users.id")),
    Column("group_id", String, ForeignKey("groups.id")),
)

user_permission = Table(
    "user_permissions",
    Base.metadata,
    Column("user_id", String, ForeignKey("users.id")),
    Column("permission_id", String, ForeignKey("permissions.id")),
)

group_permission = Table(
    "group_permissions",
    Base.metadata,
    Column("group_id", String, ForeignKey("groups.id")),
    Column("permission_id", String, ForeignKey("permissions.id")),
)



from component.schema import BaseCountSchema, BaseDateTimeSchema


class ActivityData(BaseDateTimeSchema):
    action: str
    entity: str
    data: str


class Activity(ActivityData):
    id: str

    class Config:
        from_attributes = True


class ActivityResult(BaseCountSchema):
    data: list[Activity]



from zrb.task_input.str_input import StrInput

entity_name_input = StrInput(
    name="entity-name",
    shortcut="e",
    description="Singular entity name",
    prompt="Singular entity name",
    default="book",
)

plural_entity_name_input = StrInput(
    name="plural-entity-name",
    shortcut="p",
    description="Plural entity name",
    prompt="Plural entity name",
    default=lambda m: m.get("entity_name", "") + "s",
)

main_column_name_input = StrInput(
    name="column-name",
    shortcut="c",
    description="Main column name",
    prompt="Main column name",
    default="code",
)

column_name_input = StrInput(
    name="column-name",
    shortcut="c",
    description="Column name",
    prompt="Column name",
    default="title",
)

column_type_input = StrInput(
    name="column-type",
    shortcut="t",
    description="Column type",
    prompt="Column type",
    default="str",
)



from zrb.builtin.ubuntu._group import ubuntu_group
from zrb.task_group.group import Group

ubuntu_install_group = Group(
    name="install", description="Install things on ubuntu", parent=ubuntu_group
)



import random

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.accessories.icon", attrs=["dark"]))


@typechecked
def get_random_icon() -> str:
    icons = [
        "🐶",
        "🐱",
        "🐭",
        "🐹",
        "🦊",
        "🐻",
        "🐨",
        "🐯",
        "🦁",
        "🐮",
        "🐷",
        "🍎",
        "🍐",
        "🍊",
        "🍋",
        "🍌",
        "🍉",
        "🍇",
        "🍓",
        "🍈",
        "🍒",
        "🍑",
    ]
    return random.choice(icons)



import sys

from zrb.builtin.base64 import base64_group
from zrb.builtin.devtool import devtool_group, devtool_install_group
from zrb.builtin.docker import docker_group
from zrb.builtin.env import env_group
from zrb.builtin.explain import explain_group
from zrb.builtin.git import git_group
from zrb.builtin.md5 import md5_group
from zrb.builtin.process import process_group, process_pid_group
from zrb.builtin.project import (
    project_add_app_group,
    project_add_fastapp_group,
    project_add_group,
    project_add_task_group,
    project_group,
)
from zrb.builtin.ubuntu import ubuntu_group, ubuntu_install_group
from zrb.helper.accessories.color import colored

_DEPRECATION_WARNING = """
DEPRECATED: zrb.builtin.group
User zrb.builtin instead

```python
from zrb.builtin import project_group
```
"""

print(colored(_DEPRECATION_WARNING, color="red", attrs=["bold"]), file=sys.stderr)

assert base64_group
assert devtool_group
assert devtool_install_group
assert docker_group
assert env_group
assert explain_group
assert git_group
assert md5_group
assert process_group
assert process_pid_group
assert project_group
assert project_add_group
assert project_add_app_group
assert project_add_fastapp_group
assert project_add_task_group
assert ubuntu_group
assert ubuntu_install_group



import os

from zrb import BoolInput, CmdTask, EnvFile, IntInput, StrInput, runner

from ._constant import LOAD_TEST_DIR, LOAD_TEST_TEMPLATE_ENV_FILE_NAME
from ._group import snake_zrb_app_name_group

_CURRENT_DIR = os.path.dirname(__file__)

prepare_snake_zrb_app_name_load_test = CmdTask(
    icon="🚤",
    name="prepare-load-test",
    description="Prepare load test for human readable zrb app name",
    group=snake_zrb_app_name_group,
    cwd=LOAD_TEST_DIR,
    cmd_path=[
        os.path.join(_CURRENT_DIR, "cmd", "activate-venv.sh"),
        os.path.join(_CURRENT_DIR, "cmd", "app-prepare-load-test.sh"),
    ],
)
runner.register(prepare_snake_zrb_app_name_load_test)


load_test_snake_zrb_app_name = CmdTask(
    icon="🧪",
    name="load-test",
    description="Load test human readable zrb app name",
    group=snake_zrb_app_name_group,
    upstreams=[prepare_snake_zrb_app_name_load_test],
    inputs=[
        BoolInput(
            name="kebab-zrb-app-name-load-test-headless",
            default=True,
            description="Load test UI headless",
            prompt="Load test UI headless (if True, there will be no UI)",
        ),
        IntInput(
            name="kebab-zrb-app-name-load-test-port",
            default=8089,
            description="Load test UI web port",
            prompt="Load test UI web port (Only works if headless is False)",
        ),
        IntInput(
            name="kebab-zrb-app-name-load-test-users",
            default=200,
            description="Load test users",
            prompt="Load test users",
        ),
        IntInput(
            name="kebab-zrb-app-name-load-test-spawn-rate",
            default=10,
            description="Load test spawn rate",
            prompt="Load test spawn rate",
        ),
        StrInput(
            name="kebab-zrb-app-name-load-test-url",
            default="http://localhost:zrbAppHttpPort",
            description="Load test url",
            prompt="Load test url",
        ),
    ],
    cwd=LOAD_TEST_DIR,
    env_files=[
        EnvFile(
            path=LOAD_TEST_TEMPLATE_ENV_FILE_NAME, prefix="LOAD_TEST_ZRB_ENV_PREFIX"
        )
    ],
    cmd_path=[
        os.path.join(_CURRENT_DIR, "activate-venv.sh"),
        os.path.join(_CURRENT_DIR, "app-load-test.sh"),
    ],
    should_print_cmd_result=False,
)
runner.register(load_test_snake_zrb_app_name)



from typing import Optional, TypeVar

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.string.conversion import to_cli_name
from zrb.helper.string.modification import double_quote
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask

logger.debug(colored("Loading zrb.task_group.group", attrs=["dark"]))

# flake8: noqa E501
TGroup = TypeVar("TGroup", bound="Group")


@typechecked
class Group:
    """
    Group is an umbrella of several related Tasks or sub-groups, facilitating organization and hierarchy.

    This class allows the creation of a hierarchical structure by grouping tasks and
    other task groups together. It provides methods to add tasks, retrieve tasks,
    and generate Command-Line Interface (CLI) names based on group names.

    Attributes:
        name (str): The name of the group.
        description (str): The description of the group.
        parent (Optional[TGroup]): The parent group of the current group, if any.

    Examples:
        >>> from zrb import Group
        >>> system_group = Group(name='system')
        >>> log_group = Group(name='log', parent='system')
    """

    def __init__(
        self, name: str, description: str = "", parent: Optional[TGroup] = None
    ):
        self.__name = name
        self.__description = description
        self._parent = parent
        if parent is not None:
            parent.__children.append(self)
        self.__children: list[TGroup] = []
        self.__tasks: list[AnyTask] = []

    def get_parent(self) -> Optional[TGroup]:
        """
        Retrieves parent of the Group.

        Returns:
            Optional[Group]: Parent of the group.

        Examples:
            >>> from zrb import Group
            >>> system_group = Group(name='my system')
            >>> system_log_group = Group(name='log', parent=system_group)
            >>> print(system_group.get_parent())
            >>> print(system_log_group.get_parent())
            None
            <Group "my-system">
        """
        return self._parent

    def get_description(self) -> str:
        """
        Retrieves group description.

        Returns:
            str: Description of the group.

        Examples:
            >>> from zrb import Group
            >>> group = Group(name='group', description='description of the group')
            >>> print(group.get_description())
            description of the group
        """
        return self.__description

    def get_cli_name(self) -> str:
        """
        Retrieves the CLI name of the group, formatted in kebab case.

        The method converts the group name into a CLI-friendly format, suitable for command-line usage.

        Returns:
            str: The CLI name of the group.

        Examples:
            >>> from zrb import Group
            >>> system_group = Group(name='my system')
            >>> print(system_group.get_cli_name())
            my-system
        """
        return to_cli_name(self.__name)

    def _get_full_cli_name(self) -> str:
        """
        Retrieves the full CLI name of the group, including names of parent groups.

        This method is intended for internal use and constructs a full CLI name that reflects the group's hierarchy.

        Returns:
            str: The full CLI name of the group

        Examples:
            >>> from zrb import Group
            >>> system_group = Group(name='my system')
            >>> system_log_group = Group(name='log', parent=system_group)
            >>> print(system_log_group._get_full_cli_name())
            my-system log
        """
        cli_name = self.get_cli_name()
        if self._parent is None:
            return cli_name
        parent_cli_name = self._parent._get_full_cli_name()
        return f"{parent_cli_name} {cli_name}"

    def _add_task(self, task: AnyTask):
        """
        Adds a task to the group.

        This method is intended for internal use. It appends a given task to the  group's task list.

        Args:
            task (AnyTask): The task to be added.

        Examples:
            >>> from zrb import Group, Task
            >>> first_task = Task(name='first-task')
            >>> second_task = Task(name='second-task')
            >>> group = Group(name='group')
            >>> group._add_task(first_task)
            >>> group._add_task(second_task)
            >>> print(group.get_tasks())
            [<Task "group first-task">, <Task "group second-task">]
        """
        self.__tasks.append(task)

    def get_tasks(self) -> list[AnyTask]:
        """
        Get direct Tasks under this Task Group.

        Returns:
            list[AnyTask]: list of direct Tasks under this Task Group.

        Examples:
            >>> from zrb import Group, Task
            >>> group = Group(name='group')
            >>> first_task = Task(name='first-task', group=group)
            >>> second_task = Task(name='second-task', group=group)
            >>> print(group.get_tasks())
            [<Task "group first-task">, <Task "group second-task">]
        """
        return self.__tasks

    def get_children(self) -> list[TGroup]:
        """
        Retrieves the list of direct subgroups under this group.

        Returns a list of immediate subgroups nested within this group, helping to understand the group's hierarchical structure.

        Returns:
            list[Group]: list of direct subgroups under this Task group.

        Examples:
            >>> from zrb import Group, Task
            >>> group = Group(name='group')
            >>> sub_group_1 = TaskGroup(name='sub-group-1', parent=group)
            >>> sub_group_2 = TaskGroup(name='sub-group-2', parent=group)
            >>> print(group.get_children())
            [<Group "group sub-group-1">, <Group "group sub-group-2">]
        """
        return self.__children

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        full_cli_name = double_quote(self._get_full_cli_name())
        return f"<{cls_name} {full_cli_name}>"



from integration.db_connection import engine
from integration.log import logger
from module.auth.entity.user.repo import UserDBRepo, UserRepo
from module.auth.integration.password_hasher import password_hasher

user_repo: UserRepo = UserDBRepo(
    logger=logger, engine=engine, password_hasher=password_hasher
)



from logging import Logger
from typing import Any

from component.messagebus import Publisher
from component.rpc import Caller, Server


def register_rpc(
    logger: Logger, rpc_server: Server, rpc_caller: Caller, publisher: Publisher
):
    logger.info('🥪 Register RPC handlers for "snake_zrb_module_name"')



from config import (
    APP_ENABLE_API,
    APP_ENABLE_EVENT_HANDLER,
    APP_ENABLE_LOG_MODULE,
    APP_ENABLE_RPC_SERVER,
)
from integration.app.app import app
from integration.log import logger
from integration.messagebus import consumer, publisher
from integration.rpc import rpc_caller, rpc_server
from module.auth.integration import authorizer
from module.log.api import register_api
from module.log.event import register_event
from module.log.rpc import register_rpc


def register_log():
    if not APP_ENABLE_LOG_MODULE:
        logger.info('🥪 Skip registering "log"')
        return
    if APP_ENABLE_API:
        register_api(
            logger=logger,
            app=app,
            authorizer=authorizer,
            rpc_caller=rpc_caller,
            publisher=publisher,
        )
    if APP_ENABLE_EVENT_HANDLER:
        register_event(
            logger=logger, consumer=consumer, rpc_caller=rpc_caller, publisher=publisher
        )
    if APP_ENABLE_RPC_SERVER:
        register_rpc(
            logger=logger,
            rpc_server=rpc_server,
            rpc_caller=rpc_caller,
            publisher=publisher,
        )



from zrb.builtin.project.add._group import project_add_group
from zrb.builtin.project.add.app import (
    add_app_generator,
    add_python_app,
    project_add_app_group,
)
from zrb.builtin.project.add.fastapp import (
    add_fastapp_application,
    add_fastapp_crud,
    add_fastapp_field,
    add_fastapp_module,
    project_add_fastapp_group,
)
from zrb.builtin.project.add.plugin import add_plugin
from zrb.builtin.project.add.project_task import add_project_tasks
from zrb.builtin.project.add.task import (
    add_cmd_task,
    add_docker_compose_task,
    add_python_task,
    project_add_task_group,
)

assert project_add_group
assert project_add_app_group
assert project_add_fastapp_group
assert project_add_task_group
assert add_app_generator
assert add_python_app
assert add_fastapp_application
assert add_fastapp_crud
assert add_fastapp_field
assert add_fastapp_module
assert add_plugin
assert add_project_tasks
assert add_cmd_task
assert add_docker_compose_task
assert add_python_task



import json
import os

from helper.conversion import str_to_boolean, str_to_logging_level

APP_SRC_DIR = os.path.dirname(__file__)

APP_NAME = os.getenv("APP_NAME", "app")
APP_BRAND = os.getenv("PUBLIC_BRAND", "PascalZrbAppName")
APP_TITLE = os.getenv("PUBLIC_TITLE", "PascalZrbAppName")

APP_LOGGING_LEVEL = str_to_logging_level(os.getenv("APP_LOGGING_LEVEL", "INFO"))
APP_HOST = os.getenv("APP_HOST", "0.0.0.0")
APP_PORT = int(os.getenv("APP_PORT", "8080"))
APP_RELOAD = str_to_boolean(os.getenv("APP_RELOAD", "true"))
APP_MAX_NOT_READY = int(os.getenv("APP_MAX_NOT_READY", "10"))

APP_ENABLE_OTEL = str_to_boolean(os.getenv("APP_ENABLE_OTEL", "false"))
APP_OTEL_EXPORTER_OTLP_ENDPOINT = os.getenv(
    "APP_OTEL_EXPORTER_OTLP_ENDPOINT", "http://localhost:4317"
)

APP_AUTH_ACCESS_TOKEN_COOKIE_KEY = os.getenv(
    "APP_AUTH_ACCESS_TOKEN_COOKIE_KEY", "access_token"
)
APP_AUTH_REFRESH_TOKEN_COOKIE_KEY = os.getenv(
    "APP_AUTH_REFRESH_TOKEN_COOKIE_KEY", "refresh_token"
)

APP_AUTH_ACCESS_TOKEN_TYPE = os.getenv("APP_AUTH_ACCESS_TOKEN_TYPE", "jwt")
APP_AUTH_ACCESS_TOKEN_EXPIRE_SECONDS = int(
    os.getenv("APP_AUTH_ACCESS_TOKEN_EXPIRE_SECONDS", "300")
)
APP_AUTH_REFRESH_TOKEN_TYPE = os.getenv("APP_AUTH_REFRESH_TOKEN_TYPE", "jwt")
APP_AUTH_REFRESH_TOKEN_EXPIRE_SECONDS = int(
    os.getenv("APP_AUTH_REFRESH_TOKEN_EXPIRE_SECONDS", "86400")
)
APP_AUTH_JWT_TOKEN_SECRET_KEY = os.getenv("APP_AUTH_JWT_TOKEN_SECRET_KEY", "secret")
APP_AUTH_JWT_TOKEN_ALGORITHM = os.getenv("APP_AUTH_JWT_TOKEN_ALGORITHM", "HS256")

APP_AUTH_ADMIN_ACTIVE = str_to_boolean(os.getenv("APP_AUTH_ADMIN_ACTIVE", "true"))
APP_AUTH_ADMIN_USER_ID = os.getenv("APP_AUTH_ADMIN_USER_ID", "root")
APP_AUTH_ADMIN_USERNAME = os.getenv("APP_AUTH_ADMIN_USERNAME", "admin")
APP_AUTH_ADMIN_PASSWORD = os.getenv("APP_AUTH_ADMIN_PASSWORD", "admin")
APP_AUTH_ADMIN_EMAIL = os.getenv("APP_AUTH_ADMIN_EMAIL", "")
APP_AUTH_ADMIN_PHONE = os.getenv("APP_AUTH_ADMIN_PHONE", "")
APP_AUTH_GUEST_USER_ID = os.getenv("APP_AUTH_GUEST_USER_ID", "guest")
APP_AUTH_GUEST_USERNAME = os.getenv("APP_AUTH_GUEST_USERNAME", "guest")
APP_AUTH_GUEST_EMAIL = os.getenv("APP_AUTH_GUEST_EMAIL", "")
APP_AUTH_GUEST_PHONE = os.getenv("APP_AUTH_GUEST_PHONE", "")

APP_DB_CONNECTION = os.getenv("APP_DB_CONNECTION", "sqlite://")
APP_DB_ENGINE_SHOW_LOG = str_to_boolean(os.getenv("APP_DB_ENGINE_SHOW_LOG", "false"))
APP_DB_AUTO_MIGRATE = str_to_boolean(os.getenv("APP_DB_AUTO_MIGRATE", "true"))
APP_DB_POOL_PRE_PING = str_to_boolean(os.getenv("APP_DB_POOL_PRE_PING", "true"))
APP_DB_POOL_SIZE = int(os.getenv("APP_DB_POOL_SIZE", "20"))
APP_DB_POOL_MAX_OVERFLOW = int(os.getenv("APP_DB_POOL_MAX_OVERFLOW", "0"))


APP_BROKER_TYPE = os.getenv("APP_BROKER_TYPE", "mock")

APP_ENABLE_RPC_SERVER: bool = str_to_boolean(os.getenv("APP_ENABLE_RPC_SERVER", "true"))

APP_ENABLE_EVENT_HANDLER: bool = str_to_boolean(
    os.getenv("APP_ENABLE_EVENT_HANDLER", "true")
)

APP_ENABLE_FRONTEND: bool = str_to_boolean(os.getenv("APP_ENABLE_FRONTEND", "true"))

APP_ENABLE_API: bool = str_to_boolean(os.getenv("APP_ENABLE_API", "true"))

APP_RMQ_CONNECTION_STRING = os.getenv(
    "APP_RMQ_CONNECTION", "amqp://guest:guest@localhost/"
)

APP_KAFKA_BOOTSTRAP_SERVERS = os.getenv("APP_KAFKA_BOOTSTRAP_SERVERS", "localhost:9092")
APP_KAFKA_SECURITY_PROTOCOL = os.getenv("APP_KAFKA_SECURITY_PROTOCOL", "PLAINTEXT")
APP_KAFKA_SASL_MECHANISM = os.getenv("APP_KAFKA_SASL_MECHANISM", "SCRAM-SHA-512")
APP_KAFKA_SASL_USER = os.getenv("APP_KAFKA_SASL_USER", "admin")
APP_KAFKA_SASL_PASS = os.getenv("APP_KAFKA_SASL_PASS", "admin")

APP_CORS_ALLOW_ORIGINS: list[str] = json.loads(
    os.getenv("APP_CORS_ALLOW_ORIGINS", '["*"]')
)
APP_CORS_ALLOW_ORIGIN_REGEX: str = os.getenv("APP_CORS_ALLOW_ORIGIN_REGEX", "")
APP_CORS_ALLOW_METHODS: list[str] = json.loads(
    os.getenv("APP_CORS_ALLOW_METHODS", '["*"]')
)
APP_CORS_ALLOW_HEADERS: list[str] = json.loads(
    os.getenv("APP_CORS_ALLOW_HEADERS", '["*"]')
)
APP_CORS_ALLOW_CREDENTIALS: bool = str_to_boolean(
    os.getenv("APP_CORS_ALLOW_CREDENTIALS", "false")
)
APP_CORS_EXPOSE_HEADERS: bool = str_to_boolean(
    os.getenv("APP_CORS_EXPOSE_HEADERS", "false")
)
APP_CORS_MAX_AGE: int = int(os.getenv("APP_CORS_MAX_AGE", "600"))
APP_ENABLE_AUTH_MODULE = str_to_boolean(os.getenv("APP_ENABLE_AUTH_MODULE", "true"))
APP_ENABLE_LOG_MODULE = str_to_boolean(os.getenv("APP_ENABLE_LOG_MODULE", "true"))



from collections.abc import Callable, Iterable, Mapping
from typing import Any, Optional, TypeVar, Union

from zrb.helper.accessories.color import colored
from zrb.helper.file.copy_tree import copy_tree
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import JinjaTemplate
from zrb.helper.util import (
    to_camel_case,
    to_capitalized_human_readable,
    to_human_readable,
    to_kebab_case,
    to_pascal_case,
    to_snake_case,
)
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.base_task.base_task import BaseTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.resource_maker", attrs=["dark"]))

Replacement = Mapping[str, JinjaTemplate]
ReplacementMutator = Callable[[AnyTask, Replacement], Replacement]
TResourceMaker = TypeVar("TResourceMaker", bound="ResourceMaker")


def get_default_resource_skip_parsing() -> list[str]:
    return [
        "*.mp3",
        "*.pdf",
        "*.exe",
        "*.dll",
        "*.bin",
        "*.iso",
        "*.png",
        "*.jpg",
        "*.gif",
        "*.ico",
        "*.pyc",
        "*.gz",
        "*.whl",
        "*.db",
    ]


def get_default_resource_excludes() -> list[str]:
    return [
        "*/.git",
        "*/__pycache__",
        "*/node_modules",
        "*/.venv",
        "*/poetry.lock",
        "*/.env",
    ]


@typechecked
class ResourceMaker(BaseTask):
    def __init__(
        self,
        name: str,
        template_path: JinjaTemplate,
        destination_path: JinjaTemplate,
        replacements: Replacement = {},
        replacement_mutator: Optional[ReplacementMutator] = None,
        excludes: Iterable[str] = get_default_resource_excludes(),
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        should_execute: Union[bool, JinjaTemplate, Callable[..., bool]] = True,
        skip_parsing: Iterable[str] = get_default_resource_skip_parsing(),
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=[],
            checking_interval=0.05,
            retry=0,
            retry_interval=0,
            should_execute=should_execute,
        )
        self._template_path = template_path
        self._destination_path = destination_path
        self._excludes = excludes
        self._replacements = replacements
        self._replacement_mutator = replacement_mutator
        self._skip_parsing = skip_parsing

    def copy(self) -> TResourceMaker:
        return super().copy()

    def to_function(
        self,
        env_prefix: str = "",
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True,
        should_clear_xcom: bool = False,
        should_stop_looper: bool = False,
    ) -> Callable[..., bool]:
        return super().to_function(
            env_prefix=env_prefix,
            raise_error=raise_error,
            is_async=is_async,
            show_done_info=show_done_info,
            should_clear_xcom=should_clear_xcom,
            should_stop_looper=should_stop_looper,
        )

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        # render parameters
        template_path = self.render_str(self._template_path)
        destination_path = self.render_str(self._destination_path)
        # render excludes
        self.log_debug(f"Render excludes: {self._excludes}")
        excludes = [self.render_str(exclude) for exclude in self._excludes]
        self.log_debug(f"Rendered excludes: {excludes}")
        # render replacements
        self.log_debug(f"Render replacements: {self._replacements}")
        rendered_replacements: Mapping[str, str] = {
            old: self.render_str(new) for old, new in self._replacements.items()
        }
        self.log_debug(f"Rendered replacements: {rendered_replacements}")
        if self._replacement_mutator is not None:
            self.log_debug("Apply replacement mutator")
            rendered_replacements = self._replacement_mutator(
                self, rendered_replacements
            )
        self.log_debug(f"Apply default replacement mutator: {rendered_replacements}")
        rendered_replacements = self._default_mutate_replacements(rendered_replacements)
        self.log_debug(f"Final replacement: {rendered_replacements}")
        self.print_out_dark(f"Template: {template_path}")
        self.print_out_dark(f"Destination: {destination_path}")
        self.print_out_dark(f"Replacements: {rendered_replacements}")
        self.print_out_dark(f"Excludes: {excludes}")
        self.print_out_dark(f"Skip parsing: {self._skip_parsing}")
        await copy_tree(
            src=template_path,
            dst=destination_path,
            replacements=rendered_replacements,
            excludes=excludes,
            skip_parsing=self._skip_parsing,
        )
        return True

    def _default_mutate_replacements(
        self, rendered_replacements: Mapping[str, str]
    ) -> Mapping[str, str]:
        transformations: Mapping[str, Callable[[str], str]] = {
            "Pascal": to_pascal_case,
            "camel": to_camel_case,
            "kebab": to_kebab_case,
            "KEBAB": _to_upper(to_kebab_case),
            "snake": to_snake_case,
            "SNAKE": _to_upper(to_snake_case),
            "human readable": to_human_readable,
            "Human readable": to_capitalized_human_readable,
            "HUMAN READABLE": _to_upper(to_capitalized_human_readable),
        }
        keys = list(rendered_replacements.keys())
        for key in keys:
            value = rendered_replacements[key]
            for prefix, transform in transformations.items():
                prefixed_key = transform(prefix + " " + key)
                if prefixed_key in rendered_replacements:
                    continue
                transformed_value = transform(value)
                rendered_replacements[prefixed_key] = transformed_value
            if key != key.upper():
                rendered_replacements[key.upper()] = value.upper()
        return rendered_replacements


def _to_upper(fn: Callable[[str], str]) -> Callable[[str], str]:
    def upperized_fn(text: str) -> str:
        return fn(text).upper()

    return upperized_fn



import datetime
import os
import platform
import time

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.util import (
    coalesce,
    coalesce_str,
    to_boolean,
    to_camel_case,
    to_human_readable,
    to_kebab_case,
    to_pascal_case,
    to_snake_case,
)

logger.debug(colored("Loading zrb.helper.render_data", attrs=["dark"]))

DEFAULT_RENDER_DATA = {
    "datetime": datetime,
    "os": os,
    "platform": platform,
    "time": time,
    "util": {
        "coalesce": coalesce,
        "coalesce_str": coalesce_str,
        "to_camel_case": to_camel_case,
        "to_pascal_case": to_pascal_case,
        "to_kebab_case": to_kebab_case,
        "to_snake_case": to_snake_case,
        "to_human_readable": to_human_readable,
        "to_boolean": to_boolean,
    },
}



import importlib.util
import os
import re
import sys
from functools import lru_cache

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.loader.load_script", attrs=["dark"]))

pattern = re.compile("[^a-zA-Z0-9]")


@lru_cache
@typechecked
def load_script(script_path: str, sys_path_index: int = 0):
    if not os.path.isfile(script_path):
        return
    script_dir_path = os.path.dirname(script_path)
    _append_dir_to_sys_path(script_dir_path, sys_path_index)
    _append_dir_to_python_path(script_dir_path)
    _exec_script_as_module(script_path)


def _exec_script_as_module(script_path: str):
    module_name = pattern.sub("", script_path)
    logger.info(colored(f"Get module spec: {script_path}", attrs=["dark"]))
    spec = importlib.util.spec_from_file_location(module_name, script_path)
    logger.info(colored(f"Create module: {script_path}", attrs=["dark"]))
    module = importlib.util.module_from_spec(spec)
    logger.info(colored(f"Exec module: {script_path}", attrs=["dark"]))
    spec.loader.exec_module(module)
    logger.info(colored(f"Module executed: {script_path}", attrs=["dark"]))


def _append_dir_to_sys_path(dir_path: str, index: int):
    if dir_path in sys.path:
        return
    sys.path.insert(index, dir_path)
    logger.info(colored(f"Set sys.path to {sys.path}", attrs=["dark"]))


def _append_dir_to_python_path(dir_path: str):
    new_python_path = _get_new_python_path(dir_path)
    logger.info(colored(f"Set PYTHONPATH to {new_python_path}", attrs=["dark"]))
    os.environ["PYTHONPATH"] = new_python_path


def _get_new_python_path(dir_path: str) -> str:
    current_python_path = os.getenv("PYTHONPATH")
    if current_python_path is None or current_python_path == "":
        return dir_path
    if dir_path in current_python_path.split(":"):
        return current_python_path
    return ":".join([current_python_path, dir_path])
